from datetime import datetime
import numpy as np
from ... import utils

def get_attribute_label_maps():
    metric_label_map = {"delta" : "Delta",
                        "ivol" : "Implied Volatility",
                        "IVOL_perc" : "Implied Volatility (%)",
                        "TVAR" : "Total Volatility",
                        "expiry" : "Expiry",
                        "years" : "Years",
                        "strike" : "Strike",
                        "moneyness" : "Moneyness (%)",
                        "log_moneyness" : "Log-Moneyness" ,
                        "standardised_moneyness" : "Standardised-Moneyness" 
                        }
    
    label_metric_map = {label : metric for metric, label in metric_label_map.items()}
    return metric_label_map, label_metric_map

class TickEngine:
    def __init__(self, label, valid_data=False):
        self.label=label
        self._tick_functions = {"Expiry": self.Expiry_function,
                                "Years": self._null,
                                "Delta": self.Delta_function,
                                "Strike": self._null,
                                "Moneyness (%)": self.Moneyness_func,
                                "Log-Moneyness": self._null,
                                "Implied Volatility": self._null,
                                "Total Volatility": self._null,
                                "Implied Volatility (%)" : self._null}
        
        self.function=self._tick_functions[self.label]
    
    def get_ticks(self, values, n):
        if self.label == "Delta":
            self.function()
    
    def valid_data(self, flag):
        if flag:
            self.function = self._tick_functions[self.label]
        else:
            self.function = self._null
    
    def get_label(self):
        return self.label
    
    def get_function(self):
        return self.function
    
    def change_function(self, label):
        self.label=label
        self.function=self._tick_functions[label]
        
    @staticmethod
    def _null(values):
        return [f"{value:,.2f}" for value in values]

    @staticmethod
    def Moneyness_func(values):
        if not 1 in values and (values[0] + values[-1])/2 == 1:
            values = np.append(values, 1)
            values.sort()
        return [f"{round(100 * value)}" for value in values]

    @staticmethod
    def null_visualise(values):
        return [f"{val:,.2f}" for val in values]
    
    @staticmethod
    def rounder(tick_labels):
        rounded = np.round(tick_labels)
        can_be_integer = np.all(np.isclose(tick_labels, rounded))
        if can_be_integer:
            return rounded.astype(int).astype(str).tolist()
        else:
            round_val = int(abs(np.floor(np.log10(abs(np.diff(tick_labels)).min()))))
            return np.round(tick_labels, round_val).astype(str).tolist()
    
    @staticmethod
    def Delta_function(values):
        if values[0] < 0.5 < values[-1]: 
            values = np.linspace(values[0], values[-1], len(values)).tolist()
            if not 0.5 in values:
                values = np.append(values, 0.5)
                values.sort()
            new_values=[]
            for value in values:
                if value < 50:
                    new_values.append(f"{round(value, 1)}P")
                elif value == 50:
                    new_values.append("ATM")
                else:
                    new_values.append(f"{round(1 - value, 2)}C")
            return new_values
        else:
            return [f"{round(value, 1)}P" if value < 0.5 else f"{round(1 - value, 2)}C" for value in values]
            
    @staticmethod
    def Expiry_function(values):
        return [datetime.fromtimestamp(value).strftime("%d-%b-%y").upper() if value == value and 0<value<16725186000 else str(value) for value in values]
            
    @staticmethod
    def Years_function(values):
        return [f"{val:,.2f}" for val in values]



class TickEngineManager:
    def __init__(self, x_label, y_label, z_label=None):
        self.x_engine=TickEngine(x_label)
        self.y_engine=TickEngine(y_label)
        if not z_label is None:
            self.z_engine=TickEngine(z_label)
        else:
            self.z_engine=None
    def add_valid_callbacks(self, data_container):
        data_container.add_valid_data_callback(self.x_engine.valid_data)
        data_container.add_valid_data_callback(self.y_engine.valid_data)
        if not self.z_engine is None:
            data_container.add_valid_data_callback(self.z_engine.valid_data)
    
    def change_function(self, axis_direction, axis_label):
        getattr(self, f"{axis_direction}_engine", axis_label).change_function(axis_label)
    
    def get_engine(self, axis_direction):
        return getattr(self, f"{axis_direction}_engine")
    

def get_metric_maps():    
    metric_maps = {"delta" : ["OTM", "delta", "call_flag", "delta_mag"],
                   "moneyness" : ["moneyness"],
                   "log_moneyness" : ["log_moneyness"],
                   "standardised_moneyness" : ["standardised_moneyness"],
                   "strike" : [],
                   "expiry" : [],
                   "years" : [],
                   "ivol" : [],
                   "IVOL_perc" :[],
                   "TVAR" : []}
    
    return metric_maps

def get_metric_functions():
    metric_functions = {"expiry" : null_metric,
                        "years" : years_metric_func,
                        "delta" : delta_metric_mask_sorter,
                        "strike" : null_metric,
                        "moneyness" : moneyness_mask_sorter,
                        "log_moneyness": moneyness_mask_sorter,
                        "standardised_moneyness" : moneyness_mask_sorter,
                        "ivol": null_metric,
                        "IVOL_perc" : null_metric,
                        "null" : null_metric,
                        "TVAR" : TVAR_function,
                        }
    
    return  metric_functions

def null_metric(raw_object, x, y, z):
    return x, y, z, [True]*z.size, np.arange(z.size)

def get_spot_metric_functions():
    metric_functions = get_metric_functions()
    metric_functions["moneyness"] = moneyness_spot
    metric_functions["log_moneyness"] = log_moneyness_spot
    return metric_functions

def _base_money_sorter(raw_object, x, y, z):
    mask_removal = raw_object.OTM
    x = x[mask_removal]
    y = y[mask_removal]
    z = z[mask_removal]
    
    mask_rearrange = np.lexsort((x, y))
    
    x = x[mask_rearrange]
    y = y[mask_rearrange]
    z = z[mask_rearrange]
    return x, y, z, mask_removal, mask_rearrange

def TVAR_function(raw_object, x, y, z):
    return x, y, z*utils.convert_unix_maturity_to_years(y), [True]*z.size, np.arange(z.size)

def IVOL_perc_function(raw_object, x, y, z):
    return x, y, 100*z, [True]*z.size, np.arange(z.size)

def moneyness_mask_sorter(raw_object, x, y, z):
    x=raw_object.moneyness
    return _base_money_sorter(raw_object, x, y, z)

def log_moneyness_mask_sorter(raw_object, x, y, z):
    x=raw_object.log_moneyness
    return _base_money_sorter(raw_object, x, y, z)

def standardised_moneyness_sorter(raw_object, x, y, z):
    x=raw_object.standardised_moneyness
    return _base_money_sorter(raw_object, x, y, z)

def years_metric_func(raw_object, x, y, z):
    return x, utils.convert_unix_maturity_to_years(y), z, [True]*z.size, np.arange(z.size)

def moneyness_spot(raw_object, x, y, z):
    return raw_object.moneyness, y, z, [True]*z.size, np.arange(z.size)

def log_moneyness_spot(raw_object, x, y, z):
    return raw_object.log_moneyness, y, z, [True]*z.size, np.arange(z.size)

def delta_metric_mask_sorter(raw_object, x, y, z):
    mask_removal = raw_object.OTM & (raw_object.delta_mag < 0.5)
    x = raw_object.delta
    x_masked = x[mask_removal] 
    y_masked = y[mask_removal]  
    z_masked = z[mask_removal]
    
    put_indices = np.where(~raw_object.call_flag[mask_removal])[0]
    call_indices = np.where(raw_object.call_flag[mask_removal])[0]

    if put_indices.size > 0:
        sorted_put_indices = put_indices[np.lexsort((y_masked[put_indices], -x_masked[put_indices]))]
    else:
        sorted_put_indices = np.array([], dtype=int)
    if call_indices.size > 0:
        sorted_call_indices = call_indices[np.lexsort((y_masked[call_indices], -x_masked[call_indices]))]
    else:
        sorted_call_indices = np.array([], dtype=int)

    mask_rearrange = np.concatenate([sorted_put_indices, sorted_call_indices])
    
    x_sorted = x_masked[mask_rearrange]
    y_sorted = y_masked[mask_rearrange]
    z_sorted = z_masked[mask_rearrange]        
    mask = x_sorted > 0
    x_sorted[~mask] = -1 * x_sorted[~mask]  
    x_sorted[mask] = 1 - x_sorted[mask]  
    return x_sorted, y_sorted, z_sorted, mask_removal, mask_rearrange


class MetricAxisEngine:
    def __init__(self, x_base, y_base, x_label, y_label, z_label):
        self.x_base = x_base.copy()
        self.y_base = y_base.copy()
        self.metric_label_map, self.label_metric_map = get_attribute_label_maps()      
        self.x_metric=self.label_metric_map[x_label]
        self.y_metric=self.label_metric_map[y_label]
        self.z_metric=self.label_metric_map[z_label]

        self.metric_functions = {"expiry": self.null_metric,
                                "years": self.years_metric_func,
                                "delta": self.delta_metric_mask_sorter,
                                "strike": self.null_metric,
                                "moneyness": self.moneyness_mask_sorter,
                                "log_moneyness": self.moneyness_mask_sorter,
                                "standardised_moneyness": self.moneyness_mask_sorter,
                                "ivol": self.null_metric,
                                "IVOL_perc": self.IVOL_perc_function,
                                "TVAR": self.TVAR_function,
                               }
        self.base_idx = np.arange(self.y_base.size, dtype=int)       
        
        self.x_function = self.metric_functions[self.x_metric]      
        self.y_function = self.metric_functions[self.y_metric]      
        self.z_function = self.metric_functions[self.z_metric]      
        self.x_singular = self.singular_null
        self.y_singular = self.singular_null
        self.z_singular = self.singular_null

    def singular_null(self, x):
        return x
        
    def transform_axis(self, axis_direction, args):
        return getattr(self, f"{axis_direction}_singular")(args)
    
    
    def transform_raw_object(self, raw_object):        
        x_orig, y_orig = self.x_base.copy(), self.y_base.copy()
        z_orig = raw_object.ivol.copy()

        x_metric, y, z, mask_removal_x, mask_rearrange_x = self.x_function(raw_object, x_orig, y_orig, z_orig)
        
        x_new = x_orig[mask_removal_x][mask_rearrange_x]
        
        x_new, y_metric, z, mask_removal_y, mask_rearrange_y = self.y_function(raw_object, x_new, y, z)
        
        x_metric = x_metric[mask_removal_y][mask_rearrange_y]
        y_new = y_orig[mask_removal_x][mask_rearrange_x][mask_removal_y][mask_rearrange_y]
        
        x, y, z_metric, mask_removal_z, mask_rearrange_z = self.z_function(raw_object, x_new, y_new, z)
        
        x_metric = x_metric[mask_removal_z][mask_rearrange_z]
        y_metric = y_metric[mask_removal_z][mask_rearrange_z]
        
        idx_map = self.base_idx[mask_removal_x][mask_rearrange_x][mask_removal_y][mask_rearrange_y][mask_removal_z][mask_rearrange_z]
        
        return x_metric, y_metric, z_metric, idx_map

    def get_metric(self, axis_direction):
        return getattr(self, f"{axis_direction}_metric")
    
    def get_function(self, axis_direction):
        return getattr(self, f"{axis_direction}_function")
    
    def change_function(self, axis_direction, label):
        setattr(self, f"{axis_direction}_metric", self.label_metric_map[label])
        setattr(self, f"{axis_direction}_function", self.metric_functions[self.label_metric_map[label]])
        
    
                                
    @staticmethod
    def null_metric(raw_object, x=None, y=None, z=None):
        return x, y, z, [True] * z.size, np.arange(z.size)
    
    @staticmethod
    def _base_money_sorter(raw_object, x=None, y=None, z=None):
        mask_removal = raw_object.OTM        
        if not y is None:
            x, y, z = x[mask_removal], y[mask_removal], z[mask_removal]
            mask_rearrange = np.lexsort((x, y))
            return x[mask_rearrange], y[mask_rearrange], z[mask_rearrange], mask_removal, mask_rearrange
        else:
            x, z = np.where(mask_removal, x, np.nan), np.where(mask_removal, z, np.nan)
            ind = np.argsort(x, axis=0)
            return x, None, z, mask_removal, ind

    @staticmethod
    def TVAR_function(raw_object, x=None, y=None, z=None):
        return x, y, z * utils.convert_unix_maturity_to_years(y), [True] * z.size, np.arange(z.size)
    
    @staticmethod
    def IVOL_perc_function(raw_object, x=None, y=None, z=None):
        return x, y, 100 * z, [True] * z.size, np.arange(z.size)
    
    def moneyness_mask_sorter(self, raw_object, x=None, y=None, z=None):
        x = raw_object.moneyness
        return self._base_money_sorter(raw_object, x, y, z)
    
    def log_moneyness_mask_sorter(self, raw_object, x=None, y=None, z=None):
        x = raw_object.log_moneyness
        return self._base_money_sorter(raw_object, x, y, z)
    
    def standardised_moneyness_sorter(self, raw_object, x=None, y=None, z=None):
        x = raw_object.standardised_moneyness
        return self._base_money_sorter(raw_object, x, y, z)
    
    @staticmethod
    def years_metric_func(raw_object, x=None, y=None, z=None):
        return x, utils.convert_unix_maturity_to_years(y), z, [True] * z.size, np.arange(z.size)

    def delta_metric_mask_sorter(self, raw_object, x=None, y=None, z=None, flatten=False):
        if flatten:
            mask_removal = raw_object.OTM.flatten() & (raw_object.delta_mag.flatten() < 0.5)
            put_indices = np.where(~raw_object.call_flag.flatten()[mask_removal])[0]
            call_indices = np.where(raw_object.call_flag.flatten()[mask_removal])[0]
            x_masked, y_masked, z_masked = raw_object.delta.flatten()[mask_removal], y[mask_removal], z[mask_removal]
        else:
            mask_removal = raw_object.OTM & (raw_object.delta_mag < 0.5)
            put_indices = np.where(~raw_object.call_flag[mask_removal])[0]
            call_indices = np.where(raw_object.call_flag[mask_removal])[0]
            x_masked, y_masked, z_masked = raw_object.delta[mask_removal], y[mask_removal], z[mask_removal]
        
        sorted_put_indices = put_indices[np.lexsort((y_masked[put_indices], -x_masked[put_indices]))] if put_indices.size > 0 else np.array([], dtype=int)
        sorted_call_indices = call_indices[np.lexsort((y_masked[call_indices], -x_masked[call_indices]))] if call_indices.size > 0 else np.array([], dtype=int)

        mask_rearrange = np.concatenate([sorted_put_indices, sorted_call_indices])
        x_sorted, y_sorted, z_sorted = x_masked[mask_rearrange], y_masked[mask_rearrange], z_masked[mask_rearrange]
        mask = x_sorted > 0
        x_sorted[~mask] = -x_sorted[~mask]
        x_sorted[mask] = 1 - x_sorted[mask]
        return x_sorted, y_sorted, z_sorted, mask_removal, mask_rearrange




from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, ClassVar
if TYPE_CHECKING:
    from .GL3DViewBox import GL3DViewBox 

import weakref
from pyqtgraph.Qt import isQObjectAlive
from pyvol_terminal.gl_3D_graphing.meta import QABCMeta, abc
from pyqtgraph.opengl import GLGraphicsItem
from .AbstractGLGraphicsItem import AbstractGLGraphicsItem
from .AbstractGLGraphicsItem import ABCGraphicsItemMeta
from PySide6 import QtCore, QtWidgets, QtGui
import numpy as np
from dataclasses import dataclass, field, InitVar
from abc import ABC, abstractmethod
import warnings
from pyqtgraph.Qt import QT_LIB
from functools import wraps 
import traceback
from pprint import pprint
from pyqtgraph import opengl

import importlib

if QT_LIB in ["PyQt5", "PySide2"]:
    QtOpenGL = QtGui
else:
    QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")

__all__ = ['GLLinePlotItem']




class BaseGLLinePlotItem(AbstractGLGraphicsItem, opengl.GLLinePlotItem, metaclass=ABCGraphicsItemMeta):
    def __init__(self, parentItem=None, **kwds):

        glopts = kwds.pop('glOptions', 'additive')
        self.setGLOptions(glopts)
        self.pos = None
        self.mode = 'line_strip'
        self.width = 1.
        self.color = (1.0,1.0,1.0,1.0)
        self.antialias = False

        self.m_vbo_position = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.m_vbo_color = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.vbos_uploaded = False
        
        self.setData(**kwds)

    def paintGL(self):
        return opengl.GLLinePlotItem.paint(self)


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from instruments.instruments import Option
    from pyvol_terminal.engines.surface_engines import AbstractSurfaceEngine
    from quantities.engines import MetricAxisEngine

from dataclasses import dataclass, field
from dataclasses import InitVar
import numpy as np
from typing import Any, Dict, List, Optional
import warnings
from abc import ABC, abstractmethod
import time
from PySide6 import QtCore
from .. import utils
from ..instruments.quantities import IVOLConfig


def custom_showwarning(message, category, filename, lineno, file=None, line=None):
    print(f"UserWarning: {message} ({filename}, line {lineno})")
warnings.showwarning = custom_showwarning



@dataclass(slots=True, frozen=True)
class BaseDomain:
    strike: np.ndarray
    expiry: np.ndarray
    x_metric: str
    y_metric: str
    

@dataclass
class SquaredDomain:
    base_domain: BaseDomain
    x: np.ndarray = field(init=False, default_factory=lambda: np.array)
    y: np.ndarray = field(init=False, default_factory=lambda: np.array)
    x_vect: np.ndarray = field(init=False, default_factory=lambda: np.array)
    y_vect: np.ndarray = field(init=False, default_factory=lambda: np.array)

    def __post_init__(self):
        self.x_vect = np.unique(self.base_domain.strike)
        self.y_vect = np.unique(self.base_domain.expiry)
        self.x, self.y = np.meshgrid(self.x_vect, self.y_vect, indexing="xy")

        
@dataclass(slots=True)
class Domain:
    base_domain: BaseDomain
    x: np.ndarray = field(init=False, default_factory=lambda: np.array)
    y: np.ndarray = field(init=False, default_factory=lambda: np.array)
    x_vect: np.ndarray = field(init=False, default_factory=lambda: np.array)
    y_vect: np.ndarray = field(init=False, default_factory=lambda: np.array)
    x_mat: np.ndarray = field(init=False, default_factory=lambda: np.array)
    y_mat: np.ndarray = field(init=False, default_factory=lambda: np.array)
    z_mat: np.ndarray = field(init=False, default_factory=lambda: np.array)

    xy: np.ndarray = field(init=False, default_factory=lambda: np.array)
    idx_ij_mapping: Dict[int, tuple] = field(init=False, default_factory=dict)
    z_mat: float = field(init=False, default_factory=lambda: np.array)
    
    def __post_init__(self):
        self._create_domain(self.base_domain.strike.copy(), self.base_domain.expiry.copy())

    def _create_domain(self, x, y):
        self.x = x
        self.y = y
        self.xy = np.column_stack((self.x, self.y)) 

        self.x_vect = np.unique(self.base_domain.strike)
        self.y_vect = np.unique(self.base_domain.expiry)

        self.x_mat, self.y_mat = np.meshgrid(self.x_vect, self.y_vect, indexing="xy")
        self.z_mat = np.full(self.x_mat.shape, np.nan)
        
        self.idx_ij_mapping = {}
        
        for i in range(self.x_mat.shape[0]):
            for j in range(self.x_mat.shape[1]):
                match_idx = np.where((self.xy == [self.x_mat[i, j], self.y_mat[i, j]]).all(axis=1))[0]
                for idx in match_idx:
                    self.idx_ij_mapping[idx] = (i, j)

    def update_data(self, z, idx_map):
        for idx, ivol in zip(idx_map, z):
            if idx in self.idx_ij_mapping:
                i, j = self.idx_ij_mapping[idx]
                self.z_mat[i,j] = ivol 


@dataclass(slots=True, kw_only=True)
class DomainVect:
    base_domain: InitVar[BaseDomain]
    _x: np.ndarray = field(default=None)
    _y: np.ndarray = field(default=None)
    
    def __post_init__(self, base_domain: BaseDomain):
        self.update(base_domain.strike, base_domain.expiry)

    def limits(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        return np.nanmin(self._x), np.nanmax(self._x), np.nanmin(self._y), np.nanmax(self._y)
    
    def update(self, x, y):
        self._x = x
        self._y = y
    
    def data(self):
        return self._x, self._y
    

@dataclass(slots=True)
class BaseData:
    n_instruments: int
    px_type: str
    name_index_map: Dict[str, int]
    price: np.ndarray
    
        
@dataclass(slots=True)
class RawFutures(BaseData):
    OTM: np.ndarray


@dataclass(slots=True)
class _AbstractMetrics(ABC):
    
    @abstractmethod
    def update(self, **kwargs):
        ...


@dataclass(slots=True, kw_only=True)
class Greeks(_AbstractMetrics):
    delta: np.ndarray
    gamma: np.ndarray
    vega: np.ndarray
    theta: np.ndarray
    rho: np.ndarray
    
    def update(self,
               idx,
               delta,
               gamma,
               vega,
               theta,
               rho,
               ):
        self.delta[idx] = delta
        self.gamma[idx] = gamma
        self.vega[idx] = vega
        self.theta[idx] = theta
        self.rho[idx] = rho
    

@dataclass(slots=True, kw_only=True)
class Moneyness(_AbstractMetrics):
    moneyness: np.ndarray
    log_moneyness: np.ndarray
    forward_moneyness: np.ndarray
    standardised_moneyness: np.ndarray

    def update(self,
               idx,
               moneyness,
               log_moneyness,
               forward_moneyness,
               standardised_moneyness,

               ):
        self.moneyness[idx] = moneyness
        self.log_moneyness[idx] = log_moneyness
        self.forward_moneyness[idx] = forward_moneyness
        self.standardised_moneyness[idx] = standardised_moneyness

@dataclass(slots=True, kw_only=True, frozen=True)
class Specifications:
    strike: np.ndarray
    expiry: np.ndarray
    

@dataclass(slots=True, kw_only=True)
class OptionChain(BaseData):
    ivol_config: IVOLConfig
    ivol: np.ndarray
    
    greeks: Greeks
    moneyess: Moneyness
    OTM: np.ndarray
    call_flag: np.ndarray
    valid_price: np.ndarray
    underlying_px: np.ndarray
    
    def getDataKwargs(self):
        result = {field : getattr(self, field) for field in self.__dataclass_fields__ if not field in ["greeks", "moneyness"]}
        result = result | {field : getattr(self.greeks, field) for field in self.greeks.__dataclass_fields__}
        return result | {field : getattr(self.moneyess, field) for field in self.moneyess.__dataclass_fields__}

    def update_IVOL(self, idx, implied_volatility):
        self.ivol[idx] = implied_volatility
    
    def update_from_object(self, instrument_object: Option):
        idx = self.name_index_map[instrument_object.ticker]
        self.price[idx] = getattr(instrument_object, self.px_type)
        self.ivol[idx] = instrument_object.get_ivol(self.px_type)
        self.greeks.update(idx, *instrument_object.get_greeks(self.px_type))
        self.moneyess.update(idx, *instrument_object.moneyness_metrics(self.px_type))
        self.OTM[idx] = instrument_object.OTM
        self.call_flag[idx] = instrument_object.call_flag
        self.valid_price[idx] = instrument_object.valid_price
        self.underlying_px[idx] = instrument_object.underlying_px
    
    def update_all_metrics_name(self, ticker, price, ivol, delta, gamma, vega, theta, rho, moneyness,
                                log_moneyness, standardised_moneyness, OTM, call_flag, underlying_px, valid_price):
        idx = self.name_index_map[ticker]
        self.price[idx] = price
        self.ivol[idx] = ivol
        self.greeks.update(idx, delta, gamma, vega, theta, rho)
        self.moneyess.update(idx, moneyness, log_moneyness, standardised_moneyness)
        self.OTM[idx] = OTM
        self.call_flag[idx] = call_flag
        self.valid_price[idx] = valid_price
        self.underlying_px[idx] = underlying_px
    
    def update_with_instrument_object(self, idx, option_object, ivol, update_list):
            self.ivol[idx] = ivol
            for metric in update_list:    
                getattr(self, metric)[idx] = getattr(option_object, metric)


@dataclass(slots=True, kw_only=True)
class VolVector:
    option_chain: InitVar[OptionChain]
    strikes: np.ndarray
    expiry: np.ndarray
    domain: DomainVect
    metric_engine: MetricAxisEngine
    
    _x: np.ndarray = field(init=False)
    _y: np.ndarray = field(init=False)
    _z: np.ndarray = field(init=False)
    _OTM: np.ndarray = field(init=False)
    idx_map: np.ndarray = field(default=None)
    _domain_updated_callbacks: List[Callable] = field(default_factory=list)
    
    def __post_init__(self, option_chain: OptionChain):
        self.update(**option_chain.getDataKwargs())
    
    def add_domain_updated_callback(self, callback):
        self._domain_updated_callbacks.append(callback)
    
    def update(self, **option_chain_data_kwargs):
        self._x, self._y, self._z, self.idx_map = self.metric_engine.transform_values(self.strikes,
                                                                                      self.expiry,
                                                                                      **option_chain_data_kwargs,
                                                                                      )
        self._OTM=option_chain_data_kwargs["OTM"][self.idx_map]
        self.domain.update(self._x, self._y)

    def data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return *self.domain.data(), self._z
    
    def plot_item_kwargs(self):
        return {"pos" : np.column_stack((self._x, self._y, self._z))}
    
    def calibration_args(self):
        return *self.data(), self._OTM
    
    
    


@dataclass(slots=True, kw_only=True)
class AbstractDataClass(ABC):
    domain_vect: DomainVect
    x: np.ndarray = field(default=None)
    y: np.ndarray = field(default=None)
    z: np.ndarray = field(default=None)
    
    x_min: float = field(init=False)
    x_max: float = field(init=False)
    y_min: float = field(init=False, default=None)
    y_max: float = field(init=False, default=None)
    z_min: float = np.nan
    z_max: float = np.nan
    
    valid_values: bool = False
    _update_callbacks: List[Callable] = field(default_factory=lambda: [])

    def __post_init__(self):
        self.evaluate_domain_limits(*self.domain_vect.data())
        self.z_min, self.z_max = np.nan, np.nan 

    def add_update_callback(self, callback):
        self._update_callbacks.append(callback)
    
    def remove_update_callback(self, callback):
        self._update_callbacks.remove(callback)
    
    def update_callbacks(self):
        return self._update_callbacks
    
    def evaluate_domain_limits(self, x: np.ndarray=None, y: np.ndarray=None):
        if not x is None:
            if x.size > 0:
                self.x_min, self.x_max = np.min(x), np.max(x)
            else:
                self.x_min, self.x_max = np.nan, np.nan 
        if not y is None:
            if y.size > 0:
                self.y_min, self.y_max = np.min(y), np.max(y)
            else:
                self.y_min, self.y_max = np.nan, np.nan 

    def dataRange(self):
        dataRange=[]
        for axis in "xyz":
            data = getattr(self, axis)
            data = data[~np.isnan(data)]
            if data.size == 0:
                dataRange.append(None)
            else:
                dataRange.append((data.min(), data.max()))
        return dataRange
    
    def limits(self):
        lims=[self.x_min, self.x_max]
        if not self.y_min is None:
            lims.append(self.y_min)
            lims.append(self.y_max)
        if not self.z_min is None:
            lims.append(self.z_min)
            lims.append(self.z_max)
        return tuple(lims)

    
    def get_plotitem_args(self):
        return [self.px_type, self] 
    
    def set_data2(self, x=None, y=None, z=None):
        self.x, self.y, self.z = x, y, z
        
    @abstractmethod
    def data(self):...

    @abstractmethod
    def evaluate_from_engine(self, *args):...
    
    @abstractmethod
    def plot_item_kwargs(self) -> Dict[str, Any]:...


@dataclass(slots=True, kw_only=True)
class Points(AbstractDataClass):
    
    _last_valid: np.ndarray = field(default=None)
    
    def calculate_limits(self, x, y, z):
        self.evaluate_domain_limits(x, y)
        if np.any(~np.isnan(z)):
            self.z_min, self.z_max = np.nanmin(z), np.nanmax(z)
        else:
            self.z_min, self.z_max = np.nan, np.nan
    
    def data(self)->Tuple[np.ndarray, ...]:
        return *self.domain_vect.data(), self.z
    
    def evaluate_from_engine(self, surface_engine: AbstractSurfaceEngine):
        vals = [surface_engine.evaluate([xi], [yi]) for xi, yi in zip(self.domain_vect.data()[0], self.domain_vect.data()[1])]
        try:
            z = np.array(vals)
        except:
            self.valid_values=False
            return 
        if z.size == 0:
            self.valid_values=False
        else:        
            self.z=z
            self.valid_values=True
            self._last_valid=np.column_stack(self.data())
    
    def plot_item_kwargs(self):
        x, y, z = self.data()
        
        if x.shape != z.shape or y.shape != z.shape:
            if self._last_valid is not None:
                arr=self._last_valid
            else:
                arr = np.full((1, 3), np.nan)
        else:
            arr = np.column_stack((x, y, z))
            self._last_valid=arr
        return {"pos" : arr}


@dataclass(slots=True, kw_only=True)
class Surface(AbstractDataClass):
    n_x: int
    n_y: int
    
    def __post_init__(self):   
        self.update_domain(*self.domain_vect.data())
        self.z = np.full((self.n_x, self.n_y), np.nan)

    def plot_item_kwargs(self):
        return {"x" : self.x, "y" : self.y, "z" : self.z}

    def calculate_limits(self, x, y, z):
        self.evaluate_domain_limits(x, y)
        if np.any(~np.isnan(z)):
            self.z_min, self.z_max = np.nanmin(z), np.nanmax(z)
        else:
            self.z_min, self.z_max = np.nan, np.nan

    def verify_limits(self, x_min, x_max, y_min, y_max):
        if (self.x_min != x_min
            or self.x_max != x_max
            or self.y_min != y_min
            or self.y_max != y_max):

            self.x_min = x_min 
            self.x_max = x_max
            self.y_min = y_min
            self.y_max = y_max
            self.create_interpolation_domain(x_min, x_max, y_min, y_max)
    
    def invalid_surface_cleanup(self):
        self.valid_values=False
        self.z *= np.nan
        
    def set_interpolation_domain(self, x, y):
        self.x = x
        self.y = y
    
    def update_domain(self, x, y):
        self.x_min, self.x_max = np.nanmin(x), np.nanmax(x)
        self.y_min, self.y_max = np.nanmin(y), np.nanmax(y)
        self.create_interpolation_domain(*self.limits()[:4])
    
    def calculate_interpolation_domain(self, x_min, x_max, y_min, y_max, grid_shape):
        return np.linspace(x_min, x_max, grid_shape[0]), np.linspace(y_min, y_max, grid_shape[1])

    def create_interpolation_domain(self, x_min, x_max, y_min, y_max):
        self.x, self.y = self.calculate_interpolation_domain(x_min, x_max, y_min, y_max, (self.n_x, self.n_y))
             
    def _evaluate_surface(self, surface_engine: AbstractSurfaceEngine):
        try:
            self.z = surface_engine.evaluate(self.x, self.y)
        except Exception as e:
            print(f"The interpolator could not evaluate: {e}")
            self.invalid_surface_cleanup()
            return 
        
        self.valid_values = True
        self.z_min, self.z_max = np.min(self.z), np.max(self.z)
    
    def point_data(self):
        return self.x_p, self.y_p, self.z_p
          
    def evaluate_from_engine(self, surface_engine):
        """
        if z.size < 5:
            warnings.warn(exceptions.InsufficientDataWarning(f"Unable to interpolate: z has only {z.size} points (min 5 required)", 
                                                            code=100),
                        stacklevel=2  # Point to the caller's line
                        )
            _ = self.invalid_surface_cleanup()
        else:
        """
        self.verify_limits(*self.domain_vect.limits())
        self._evaluate_surface(surface_engine)
        if self.valid_values:
            for callback in self._update_callbacks:
                callback(self)
                    
    @classmethod
    def from_points(cls,
                    points: Points,
                    n_ticks: Tuple[int, int],
                   ) -> 'Surface':
        x, y = cls.calculate_interpolation_domain(cls, *points.limits()[:4], n_ticks)
        return cls(x=x,
                   y=y,
                   x_p=points.x,
                   y_p=points.y,
                   z_p=points.z,
                   n_x=n_ticks[0],
                   n_y=n_ticks[1],
                   )
        
    def data(self)->Tuple[np.ndarray, ...]:
        return self.x, self.y, self.z


@dataclass(slots=True, kw_only=True)
class VolatilityData:
    px_type: str
    
    vol_vector: VolVector
    scatter: Points = field(default=None)
    surface: Surface = field(default=None)
    skew: Slice = field(default=None)
    term: Slice = field(default=None)   
    _last_update: float = field(default_factory=lambda: time.time())
    
    def surface_children(self) -> Tuple[Slice, Slice]:
        return self.skew, self.term
    
    def get_dataclass(self, plot_type):
        match plot_type:
            case "surface":
                return self.surface
            case "scatter":
                return self.scatter
            case "skew":
                return self.skew
            case "term":
                return self.term



@dataclass(slots=True, kw_only=True)
class Slice(AbstractDataClass):
    #domain_vect=DomainVect
    axis_3D_directions: str
    name: str = field(default=None)
    
    xi: np.ndarray = field(default_factory=lambda: np.array([]))
    yi: np.ndarray = field(default_factory=lambda: np.array([]))
    
    perp_x_point: float|None = field(default=None)
    
    _last_engine: AbstractSurfaceEngine|None = field(default=None)
    _last_surface: Tuple[np.ndarray, np.ndarray, np.ndarray] = field(default=None)
    _qt_object: QtCore.QObject = field(default_factory=QtCore.QObject)
    _last_point: Tuple[float, float]|None = field(default=None)
    
    def update_from_surface(self,
                            surface: Surface
                            ):
        self._last_surface=surface.data()
        if not self._last_point is None:
            self.update_data(*self._last_surface,
                             *self._last_point)
        
    def set_perpendicular_x(self, point: float):
        self.perp_x_point=point
        self._update_point_axis()
    
    def _update_point_axis(self):
        xi, yi = self.domain_vect.data()
        if self.axis_3D_directions=="xz": 
            self.yi = [self.perp_x_point] * xi.size
        else:
            self.xi = [self.perp_x_point] * yi.size

    def update_domain(self, x, y):
        if self.axis_3D_directions=="xz": 
            self.x=x
            self.yi = [self.perp_x_point] * x.size
            self.xi=x
        else:
            self.x = y
            self.xi = [self.perp_x_point] * y.size
            self.yi=x
        
    def plot_item_kwargs(self):
        return {"x" : self. x, "y" : self.y}
        
    def update_point(self, *args):
        self._last_point = args[0], args[1]
        if not self._last_surface is None:
            self.update_data(*self._last_surface,
                             *self._last_point)
            for callback in self._update_callbacks:
                callback(self)
        
    def update_data(self,
                    x_surf: np.ndarray,
                    y_surf: np.ndarray,
                    z_surf: np.ndarray,
                    x_p: float,
                    y_p: float
                    ):
        
        line = utils.calculate_xy_lines(x_surf,
                                        y_surf,
                                        z_surf,
                                        x_p,
                                        y_p,
                                        self.axis_3D_directions=="xz"
                                        )
        if self.axis_3D_directions=="xz":
            self.x = line[:,0]
        else:
            self.x = line[:,1]
        
        self.y = line[:,2]
        self.valid_values=True
        
    def data(self):
        return self.x, self.y

    def connect(self, *args, **kwargs):
        return self._qt_object.connect(*args, **kwargs)
    
    def disconnect(self, *args, **kwargs):
        return self._qt_object.disconnect(*args, **kwargs)
    
    def evaluate_from_engine(self, *args):
        return            
    

@dataclass(slots=True, kw_only=True)
class SliceContainer:
    dependent_axis: int
    points: Points
    
    def get_slice(self, point: float):
        pass
            

@dataclass(slots=True, kw_only=True)
class Slices222(AbstractDataClass):
    surface: Surface = None
    surface_axis_directions: str = ""
    base_domain: InitVar[BaseDomain] = None
    scatter: Optional[Points] = field(init=True)
    displayed_slices: List[float] = field(init=False)
    slice_container: Dict[float, List[np.ndarray]] = field(init=False)
    n_x: int = field(init=False)
    domain_vec: np.ndarray = field(init=False)
    domain_idx_map: Dict[str, int] = field(init=False)
        
    def __post_init__(self, base_domain):
        if self.surface_axis_directions == "xz":
            self.domain_vec = np.unique(base_domain.expiry)
            domain_matrix = self.x
            self.n_x = self.surface.y.size
        else:
            self.domain_vec = np.unique(base_domain.strike)
            domain_matrix = self.y
            self.n_x = self.surface.x.size
            
        if self.domain_vec.size > 15:
            self.filter_domain(base_domain)

        self.domain_idx_map = {val: idx for idx, val in enumerate(self.domain_vec)}
        self.slice_container={}

        for x_val in self.domain_vec:
            self.slice_container[x_val] = np.full(self.n_x, np.nan)
        
    def filter_domain(self, base_domain):
        value_dict = {}
        for value in self.domain_vec:
            if self.surface_axis_directions == "xz":
                mask = base_domain.expiry == value
                filtered_domain = base_domain.expiry[mask]
            else:
                mask = base_domain.strike == value
                filtered_domain = base_domain.strike[mask]
            value_dict[value] = len(filtered_domain)
        top_12 = sorted(value_dict, key=value_dict.get, reverse=True)[:12]
        top_12 = np.sort(top_12)
        self.domain_vec = self.domain_vec[np.isin(self.domain_vec, top_12)]

    def evaluate_from_engine(self, x, y, z):
        if not self.scatter is None:
            self.scatter.evaluate_from_engine(x, y, z)
        self.surface.evaluate_from_engine(x, y, z)

        if not self.valid_values:
            return 
        else:
            self.update_slice(x, y, z)
    
    def _interpolate_slice(self, dependent_values, constant_values):
        if self.surface_axis_directions == "xz":
            z_interp = self.interpolator.evaluate(dependent_values, constant_values)
        else:
            z_interp = self.interpolator.evaluate(constant_values, dependent_values)

        if z_interp.ndim != 1:
            if z_interp.shape[1] == 1 or z_interp.shape[0] == 1:
                z_interp = z_interp.flatten()
            else:
                if self.surface_axis_directions=="yz":
                    z_interp = z_interp[0,:]
                else:
                    z_interp = z_interp[:,0]
        return z_interp

    def update_slice(self, x, y, z):
        dependent_values = x if self.surface_axis_directions == "xz" else y
        size = dependent_values.size
        constant_values = [np.nan] * size
        for constant_value in self.displayed_slices:
            constant_values[:] = [constant_value] * size
            z_interp = self._interpolate_slice(dependent_values, constant_values)        



from PySide6 import QtWidgets, QtCore
from datetime import datetime   
import numpy as np

class VolTable(QtWidgets.QTableWidget):
    def __init__(self, data_container_manager, tick_label_engine=None, parent=None):
        self.data_container_manager=data_container_manager
        self.tick_label_engine=tick_label_engine
        
        self.column_items=[]
        self.row_items=[]
        
        self.init_xy_data(data_container_manager)

        super().__init__(self.rows, self.columns, parent)
        self.setHorizontalHeaderLabels(self.column_vals)
        self.setVerticalHeaderLabels(self.row_vals)
        self.update_table()


    def init_xy_data(self, data_container_manager):
        self.domain_mid = data_container_manager.objects["mid"].domain
        
        self.row_vals = self.domain_mid.y_vect
        self.column_vals = self.domain_mid.x_vect
        self.rows = self.row_vals.size
        self.columns = self.column_vals.size
        
        self.column_vals=[self.tick_label_engine.x_func(new_val) for new_val in self.column_vals]
        self.row_vals=[self.tick_label_engine.y_func(new_val) for new_val in self.row_vals]

    def update_table(self):
        self.blockSignals(True)
        self.setUpdatesEnabled(False)        
        for idx in range(self.domain_mid.z_mat.shape[0]):
            for jdx in range(self.domain_mid.z_mat.shape[1]):
                new_val_str = self.tick_label_engine.z_func(self.domain_mid.z_mat[idx,jdx])
                item = QtWidgets.QTableWidgetItem(new_val_str)
                self.setItem(idx, jdx, item)
        
        self.blockSignals(False)
        self.setUpdatesEnabled(True)
    
    def _update_table_labels(self):
        self.row_vals = self.domain_mid.y_vect
        self.column_vals = self.domain_mid.x_vect
        self.rows = self.row_vals.size
        self.columns = self.column_vals.size
        
        self.setColumnCount(self.columns)
        self.setRowCount(self.rows)
        
        new_cols=[self.tick_label_engine.x_func(new_val) for new_val in self.column_vals]
        self.setHorizontalHeaderLabels(new_cols)
        
        new_rows=[self.tick_label_engine.y_func(new_val) for new_val in self.row_vals]
        self.setVerticalHeaderLabels(new_rows)




class CustomQTableWidgetItem(QtWidgets.QTableWidgetItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
    

class OptionMonitorTable(QtWidgets.QWidget):
    def __init__(self, instrument_manager, parent=None):
        super().__init__(parent)
        self.instrument_manager=instrument_manager
        self.KT_instrument_obj_map = {}
        self.instrument_obj_textitem_maps = {}        
        self.data_cols = ['instrument_name', 'bid', 'ask', 'mid', 'IVOL']
        
        self.v_layout = QtWidgets.QVBoxLayout(self)
        
        
        self.n_data_cols = len(self.data_cols)
        self.metric_columns = self.data_cols[1:]
        self.child1_n_columns = len(self.data_cols)
        self.setup_parent_table(self.data_cols, 5)
        self.expiry_strike_map = {exp : self.strikes for exp in self.expiry}
        
        self.initialise_child_tables(self.option_table)
        self.header_layout.setColumnCount(self.option_table.columnCount())
        for col in range(self.option_table.columnCount()):
            self.header_layout.setColumnWidth(col, self.option_table.columnWidth(col))
            


    def _create_table_maps(self, n_strikes):
        strikes = []
        expiry = []
        for instrument_name, option_object in self.instrument_manager.options.items():
            strikes.append(option_object.strike)
            expiry.append(option_object.expiry)

        self.expiry=np.unique(expiry)
        self.strikes=np.unique(strikes)
        
        centre = 0.5 * (self.strikes.min() + self.strikes.max())
        differences = np.abs(centre - self.strikes)
        closest_indices  = np.argsort(differences)[:n_strikes]
        
        self.strikes = self.strikes[closest_indices]
        
        self.expiry_strike_map = {}
        
        self.expiry.sort()
        self.strikes.sort()
        
        self.KT_instrument_obj_map = {}

        for T in self.expiry:
            for K in self.strikes:
                for instrument_name, option_object in self.instrument_manager.options.items():
                    if T == option_object.expiry and K == option_object.strike:
                        if option_object.flag_int == 1:
                            call_obj = option_object
                            put_name = self.instrument_manager.options_maps.put_call_map[instrument_name]
                            put_obj = self.instrument_manager.options[put_name]
                            break
                        else:
                            put_obj = option_object
                            call_name = self.instrument_manager.options_maps.put_call_map[instrument_name]
                            call_obj = self.instrument_manager.options[call_name]

                            break
                        
                self.KT_instrument_obj_map[(K,T)] = {"call" : call_obj,
                                                     "put" : put_obj}
                if "call" in self.KT_instrument_obj_map[(K,T)] and "put" in self.KT_instrument_obj_map[(K,T)]:
                    self.expiry_strike_map[T] = self.strikes

    def _create_header_tables(self, data_cols):
        self.header_layout = QtWidgets.QHBoxLayout()

        self.header_layout.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, 
            QtWidgets.QSizePolicy.Minimum  
        )

    
        n_data_cols = len(data_cols)

        n_cols = 1 + int(2 * n_data_cols)

        self.header_layout.setColumnCount(n_cols)
        self.header_layout.setRowCount(1)
        self.header_layout.setSpan(0, 0, 1, n_data_cols)
        self.header_layout.setSpan(0, n_data_cols + 1, 1, n_data_cols)
        h_c = QtWidgets.QLabel("Calls")
        h_s = QtWidgets.QLabel("Strikes")
        h_p = QtWidgets.QLabel("Puts")
        
        self.header_layout.addWidget(h_c)
        self.header_layout.addWidget(h_s)
        self.header_layout.addWidget(h_p)
        
        return self.header_layout
    
    def setup_parent_table(self, data_cols, n_strikes):
        
        self.header_layout=self._create_header_tables(data_cols)
        self._create_table_maps(n_strikes)
        self.option_table = QtWidgets.QTableWidget()
        self.option_table.verticalHeader().setVisible(False)
        self.option_table.horizontalHeader().setVisible(True)
        n_rows = len(self.expiry) * (1 + n_strikes) 
        self.option_table.setRowCount(n_rows)
        
        self.header_layout.verticalHeader().setVisible(False)
        self.header_layout.horizontalHeader().setVisible(False)
        self.header_layout.setSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)

        self.header_layout
        
        self._create_table_maps(n_strikes)

        n_rows = int(n_strikes * (len(self.expiry + 1)))
        
        self.option_table.setColumnCount(self.header_layout.columnCount())
        self.option_table.setRowCount(n_rows)
        
        self.option_table.setHorizontalHeaderLabels(data_cols  + [""] + data_cols)
        self.v_layout.addWidget(self.header_layout)
        self.v_layout.addWidget(self.option_table)
        

        
    def update_table(self):
        self.blockSignals(True)
        self.setUpdatesEnabled(False)        
        for instrument_obj, text_item_dict in self.instrument_obj_textitem_maps.items():
            for metric, text_item in text_item_dict.items():
                val = getattr(instrument_obj, metric)
                if metric == "IVOL":
                    val = val[2]

                text_item.setText(str(np.round(val, 2)))
        self.blockSignals(False)
        self.setUpdatesEnabled(True)     
           

    def initialise_child_tables(self, option_table):
        idx = 0
        idx_sets_c = []
        idx_sets_p = []
        for exp, strikes in self.expiry_strike_map.items():
            
            item_call = QtWidgets.QTableWidgetItem(datetime.fromtimestamp(exp).strftime("%d-%b-%y") + "c")
            
            item_call.setTextAlignment(QtCore.Qt.AlignLeft)
            font_call = item_call.font()
            font_call.setPointSize(12)
            item_call.setFont(font_call)
            item_call.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
            
            item_put = CustomQTableWidgetItem(datetime.fromtimestamp(exp).strftime("%d-%b-%y") + 'p')
            item_put.setTextAlignment(QtCore.Qt.AlignLeft)
            font_put = item_put.font()
            font_put.setPointSize(12)
            item_put.setFont(font_put)
            item_put.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
            
            option_table.setSpan(idx, 0, 1, self.n_data_cols) 
            option_table.setSpan(idx, self.n_data_cols + 1, 1, self.n_data_cols)
            
            line_edit_widget = self.create_line_edit(len(strikes))
            
            option_table.setItem(idx, 0, item_call)
            option_table.setCellWidget(idx, self.n_data_cols, line_edit_widget)
            option_table.setItem(idx, self.n_data_cols + 1, item_put)
            
            idx += 1
        
            for strike in strikes:
                call_object = self.KT_instrument_obj_map[(strike, exp)]["call"]
                put_object = self.KT_instrument_obj_map[(strike, exp)]["put"]

                item_c = CustomQTableWidgetItem(str(call_object.instrument_name))
                item_p = CustomQTableWidgetItem(str(put_object.instrument_name))

                item_c.setTextAlignment(QtCore.Qt.AlignLeft)
                item_p.setTextAlignment(QtCore.Qt.AlignLeft)
                
                if (idx,0) in idx_sets_c:
                    print("call")
                    print(idx, 0)
                    
                if (idx,0) in idx_sets_p:
                    print("put")
                    print(idx, 0)
            
                item_s = CustomQTableWidgetItem(str(strike))
                
                option_table.setItem(idx, 0, item_c)
                option_table.setItem(idx, self.n_data_cols, item_s)
                option_table.setItem(idx, self.n_data_cols + 1, item_p)
                
                idx_sets_c.append((idx,0))
                idx_sets_p.append((idx,0))

                self.instrument_obj_textitem_maps[call_object] = item_c
                self.instrument_obj_textitem_maps[put_object] = item_p

                call_metric_dict = {}
                put_metric_dict = {}
                jdx = 1
                
                for metric in self.metric_columns:
                    c_val = getattr(call_object, metric)
                    if metric == "IVOL":
                        c_val = c_val[2]
                    if c_val != c_val:
                        c_val=str(c_val)
                    else:
                        c_val=str(np.round(c_val),2)
                    
                    p_val = getattr(put_object, metric)
                    if metric == "IVOL":
                        p_val = p_val[2]

                    if p_val != p_val:
                        p_val=str(p_val)
                    else:
                        p_val=str(np.round(p_val),2)
                        
                        
                        
                    item_c = CustomQTableWidgetItem(c_val)
                    item_p = CustomQTableWidgetItem(p_val)
                    call_metric_dict[metric] = item_c
                    put_metric_dict[metric] = item_p

                    item_c.setTextAlignment(QtCore.Qt.AlignCenter)
                    item_p.setTextAlignment(QtCore.Qt.AlignCenter)
                    
                    if (idx,jdx) in idx_sets_c:
                        print("call")
                        print(idx, jdx)
                        
                    if (idx,jdx) in idx_sets_p:
                        print("put")
                        print(idx, jdx)

                    idx_sets_c.append((idx,jdx))
                    idx_sets_p.append((idx,jdx))


                    option_table.setItem(idx, jdx, item_c)
                    option_table.setItem(idx, jdx + self.n_data_cols + 1, item_p)
                    jdx+=1
                    
                idx+=1
                self.instrument_obj_textitem_maps[call_object] = call_metric_dict
                self.instrument_obj_textitem_maps[put_object] = put_metric_dict

                
    def create_line_edit(self, n_strikes):
        line_edit = QtWidgets.QLineEdit()
        line_edit.setText(f"{n_strikes}") 
        line_edit.setAlignment(QtCore.Qt.AlignCenter)  
        line_edit.setStyleSheet("""
                                QLineEdit {
                                        background-color: white;
                                        color: black;
                                        text-align: center;
                                 }
                                    QLineEdit:focus {
                                        background-color: #ffffcc; 
                                        border: 2px solid #3366ff;
                                 }
                                """)
        line_edit.editingFinished.connect(self.change_strikes)
        return line_edit
    
    def change_strikes(self):
        line_edit = self.sender() 
        typed_text = line_edit.text() 
        n_strikes = int(float(typed_text))
        line_edit.clearFocus()
        



import sys
from PySide6.QtWidgets import QApplication, QMainWindow, QMenuBar, QToolBar
from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon, QAction
import sys


class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        
        # Create menu bar
        menubar = self.menuBar()
        
        # Create File menu
        file_menu = menubar.addMenu("File")
        
        # Create actions with different display text and icon text
        action1 = QAction(QIcon.fromTheme("document-new"), "New Document", self)
        action1.setIconText("New")  # Shorter text for icon display
        
        action2 = QAction(QIcon.fromTheme("document-open"), "Open Existing File...", self)
        action2.setIconText("Open")  # Shorter text for icon display
        
        action3 = QAction(QIcon.fromTheme("document-save"), "Save Current Document", self)
        action3.setIconText("Save")  # Shorter text for icon display
        
        # Add actions to menu
        file_menu.addAction(action1)
        file_menu.addAction(action2)
        file_menu.addAction(action3)
        
        # Create toolbar and force it to show text
        toolbar = QToolBar("Main Toolbar")
        toolbar.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)  # Show text under icons
        self.addToolBar(toolbar)
        
        # Add actions to toolbar
        toolbar.addAction(action1)
        toolbar.addAction(action2)
        toolbar.addAction(action3)
        
        # Connect actions to slots
        action1.triggered.connect(lambda: print("New action triggered"))
        action2.triggered.connect(lambda: print("Open action triggered"))
        action3.triggered.connect(lambda: print("Save action triggered"))
        
        self.setWindowTitle("QAction Text Display Demo")
        self.setGeometry(100, 100, 400, 300)

if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec_())


#%%

import py_vollib_vectorized
import numpy as np

def generate_synthetic_surface(market_params, S, K, T):
    ATM_vol = market_params["ATM_vol"]
    kurtosis_1 = market_params["kurtosis_1"]
    kurtosis_2 = market_params["kurtosis_2"]
    decay = market_params["decay"]
    
    if K < S:
        smile=ATM_vol + kurtosis_1 * ((K - S) / S) ** 2
    else:
        smile=ATM_vol + kurtosis_2 * ((K - S) / S) ** 2
    
    flattening_factor = np.exp(-decay * T)
    IV = smile * flattening_factor
    return IV


vol_params = {"ATM_vol" : 0.20,
                "kurtosis_1" : 0.55,
                "kurtosis_2" : 0.2,
                "decay" : 0.1
                }


def generate_spread_pct(price, min_spread=0.001, max_spread=0.05):
    inv_price = 1 / price
    norm_inv = (inv_price - inv_price.min()) / (inv_price.max() - inv_price.min())
    return min_spread + norm_inv * (max_spread - min_spread)

mid_prices = np.linspace(5, 30, 30)[::-1]

spread_pcts = generate_spread_pct(mid_prices)
random_noise = np.random.uniform(0.9, 1.1, size=len(mid_prices))
spread_pcts *= random_noise
bids = mid_prices * (1 - spread_pcts / 2)
asks = mid_prices * (1 + spread_pcts / 2)


S = 130.
K = np.linspace(100, 130, 30)
K = [100.]
prices = np.arange(5, 200, dtype=float)
t = 1
r = 0
flag = ['c']
vols=[]
for p in prices:
    vol=100 * py_vollib_vectorized.vectorized_implied_volatility(p, S, K, t, r, flag, q=0, model='black_scholes_merton',return_as='numpy').item()
    vols.append(vol)
import matplotlib.pyplot as plt
plt.plot(prices, vols)
plt.xlabel("Price ($)")
plt.ylabel("IVOL (%)")
plt.show()
#%%
vols_b=[]
vols_a=[]

vols=[]
for k in K:
    vol = generate_synthetic_surface(vol_params, S, k, t)
    vols.append(vol)


for b, a, k in zip(bids, asks, K):
    ivol_b = py_vollib_vectorized.vectorized_implied_volatility(b, S, k, t, r, flag, q=0, model='black_scholes_merton',return_as='numpy').item()
    ivol_a = py_vollib_vectorized.vectorized_implied_volatility(a, S, k, t, r, flag, q=0, model='black_scholes_merton',return_as='numpy').item()
    vols_b.append(ivol_b)
    vols_a.append(ivol_a)

#%%
from scipy.interpolate import make_interp_spline
from scipy.stats import linregress
m = (vols[1] - vols[0]) / (prices[1] - prices[0])

c = vols[0] - m * prices[0] 
slope, intercept, r_value, p_value, std_err = linregress(prices, vols)
spline = make_interp_spline(prices, vols, k=1)

#%%
import matplotlib.pyplot as plt
plt.plot(K, vols_b)
plt.plot(K, vols_a)

plt.show()


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from .GLViewBox import GLViewBox 

#from pyqtgraph.opengl import GLGraphicsItem
import weakref
from pyqtgraph.Qt import QtCore, QtGui, QtWidgets, isQObjectAlive
import operator
from functools import reduce
from abc import ABCMeta, abstractmethod
import numpy as np
from ..import meta
from pyqtgraph import Transform3D
from pyqtgraph import functions as fn
import abc
import traceback
import warnings
from pyqtgraph.Qt import QT_LIB
from .GLGraphicsItem import GLGraphicsItem
from PySide6.QtWidgets import QGraphicsItem


class EnumContainer:
    NoCache = QGraphicsItem.CacheMode.NoCache
    ItemCoordinateCache = QGraphicsItem.CacheMode.ItemCoordinateCache
    DeviceCoordinateCache = QGraphicsItem.CacheMode.DeviceCoordinateCache
    
    ItemEnabledChange = QGraphicsItem.GraphicsItemChange.ItemEnabledChange
    ItemEnabledHasChanged = QGraphicsItem.GraphicsItemChange.ItemEnabledHasChanged
    ItemPositionChange = QGraphicsItem.GraphicsItemChange.ItemPositionChange
    ItemPositionHasChanged = QGraphicsItem.GraphicsItemChange.ItemPositionHasChanged
    ItemTransformChange = QGraphicsItem.GraphicsItemChange.ItemTransformChange
    ItemTransformHasChanged = QGraphicsItem.GraphicsItemChange.ItemTransformHasChanged
    ItemRotationChange = QGraphicsItem.GraphicsItemChange.ItemRotationChange
    ItemRotationHasChanged = QGraphicsItem.GraphicsItemChange.ItemRotationHasChanged
    ItemScaleChange = QGraphicsItem.GraphicsItemChange.ItemScaleChange
    ItemScaleHasChanged = QGraphicsItem.GraphicsItemChange.ItemScaleHasChanged
    ItemTransformOriginPointChange = QGraphicsItem.GraphicsItemChange.ItemTransformOriginPointChange
    ItemTransformOriginPointHasChanged = QGraphicsItem.GraphicsItemChange.ItemTransformOriginPointHasChanged
    ItemSelectedChange = QGraphicsItem.GraphicsItemChange.ItemSelectedChange
    ItemSelectedHasChanged = QGraphicsItem.GraphicsItemChange.ItemSelectedHasChanged
    ItemVisibleChange = QGraphicsItem.GraphicsItemChange.ItemVisibleChange
    ItemVisibleHasChanged = QGraphicsItem.GraphicsItemChange.ItemVisibleHasChanged
    ItemParentChange = QGraphicsItem.GraphicsItemChange.ItemParentChange
    ItemParentHasChanged = QGraphicsItem.GraphicsItemChange.ItemParentHasChanged
    ItemChildAddedChange = QGraphicsItem.GraphicsItemChange.ItemChildAddedChange
    ItemChildRemovedChange = QGraphicsItem.GraphicsItemChange.ItemChildRemovedChange
    ItemSceneChange = QGraphicsItem.GraphicsItemChange.ItemSceneChange
    ItemSceneHasChanged = QGraphicsItem.GraphicsItemChange.ItemSceneHasChanged
    ItemCursorChange = QGraphicsItem.GraphicsItemChange.ItemCursorChange
    ItemCursorHasChanged = QGraphicsItem.GraphicsItemChange.ItemCursorHasChanged
    ItemToolTipChange = QGraphicsItem.GraphicsItemChange.ItemToolTipChange
    ItemToolTipHasChanged = QGraphicsItem.GraphicsItemChange.ItemToolTipHasChanged
    ItemFlagsChange = QGraphicsItem.GraphicsItemChange.ItemFlagsChange
    ItemFlagsHaveChanged = QGraphicsItem.GraphicsItemChange.ItemFlagsHaveChanged
    ItemZValueChange = QGraphicsItem.GraphicsItemChange.ItemZValueChange
    ItemZValueHasChanged = QGraphicsItem.GraphicsItemChange.ItemZValueHasChanged
    ItemOpacityChange = QGraphicsItem.GraphicsItemChange.ItemOpacityChange
    ItemOpacityHasChanged = QGraphicsItem.GraphicsItemChange.ItemOpacityHasChanged
    ItemScenePositionHasChanged = QGraphicsItem.GraphicsItemChange.ItemScenePositionHasChanged

    ItemIsMovable = QGraphicsItem.GraphicsItemFlag.ItemIsMovable.value
    ItemIsSelectable = QGraphicsItem.GraphicsItemFlag.ItemIsSelectable.value
    ItemIsFocusable = QGraphicsItem.GraphicsItemFlag.ItemIsFocusable.value
    ItemClipsToShape = QGraphicsItem.GraphicsItemFlag.ItemClipsToShape.value
    ItemClipsChildrenToShape = QGraphicsItem.GraphicsItemFlag.ItemClipsChildrenToShape.value
    ItemIgnoresTransformations = QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations.value
    ItemIgnoresParentOpacity = QGraphicsItem.GraphicsItemFlag.ItemIgnoresParentOpacity.value
    ItemDoesntPropagateOpacityToChildren = QGraphicsItem.GraphicsItemFlag.ItemDoesntPropagateOpacityToChildren.value
    ItemStacksBehindParent = QGraphicsItem.GraphicsItemFlag.ItemStacksBehindParent.value
    ItemUsesExtendedStyleOption = QGraphicsItem.GraphicsItemFlag.ItemUsesExtendedStyleOption.value
    ItemHasNoContents = QGraphicsItem.GraphicsItemFlag.ItemHasNoContents.value
    ItemSendsGeometryChanges = QGraphicsItem.GraphicsItemFlag.ItemSendsGeometryChanges.value
    ItemAcceptsInputMethod = QGraphicsItem.GraphicsItemFlag.ItemAcceptsInputMethod.value
    ItemNegativeZStacksBehindParent = QGraphicsItem.GraphicsItemFlag.ItemNegativeZStacksBehindParent.value
    ItemIsPanel = QGraphicsItem.GraphicsItemFlag.ItemIsPanel.value
    ItemSendsScenePositionChanges = QGraphicsItem.GraphicsItemFlag.ItemSendsScenePositionChanges.value
    ItemContainsChildrenInShape = QGraphicsItem.GraphicsItemFlag.ItemContainsChildrenInShape.value


class _BaseAbstractGraphicsItem(QtCore.QObject, EnumContainer, GLGraphicsItem):
    itemChanged = QtCore.Signal(QGraphicsItem.GraphicsItemChange, object)
    def __init__(self, parentItem=None):
        super().__init__(parentItem)
        self.__inform_view_on_changes=True
        self._cache_mode = self.NoCache
        self._flags = 0
        self._position = (0, 0)
        self._visible = True
        self._enabled = True
        self._selected = False
        self._transform = None
        self._rotation = 0.0
        self._scale = 1.0
        self._transform_origin = (0, 0)
        self._z_value = 0.0
        self._opacity = 1.0
        
        

    
    def itemChange(self, change, value):
        ret = super().itemChange(change, value)

        if change in [self.GraphicsItemChange.ItemParentHasChanged, self.GraphicsItemChange.ItemSceneHasChanged]:
            if self.__class__.__dict__.get('parentChanged') is not None:
                # user's GraphicsObject subclass has a parentChanged() method
                warnings.warn(
                    "parentChanged() is deprecated and will be removed in the future. "
                    "Use changeParent() instead.",
                    DeprecationWarning, stacklevel=2
                )
                if QT_LIB == 'PySide6' and QtCore.__version_info__ == (6, 2, 2):
                    # workaround PySide6 6.2.2 issue https://bugreports.qt.io/browse/PYSIDE-1730
                    # note that the bug exists also in PySide6 6.2.2.1 / Qt 6.2.2
                    getattr(self.__class__, 'parentChanged')(self)
                else:
                    self.parentChanged()
            else:
                self.changeParent()
        try:
            inform_view_on_change = self.__inform_view_on_changes
        except AttributeError:
            # It's possible that the attribute was already collected when the itemChange happened
            # (if it was triggered during the gc of the object).
            pass
        else:
            if inform_view_on_change and change in [self.GraphicsItemChange.ItemPositionHasChanged, self.GraphicsItemChange.ItemTransformHasChanged]:
                self.informViewBoundsChanged()
            
        return ret

        
    def itemChange(self, change, value):

        self.itemChanged.emit(change, value)
        return value
    
    # Cache mode property
    @property
    def cacheMode(self):
        return self._cache_mode
    
    @cacheMode.setter
    def cacheMode(self, value):
        if self._cache_mode != value:
            self._cache_mode = value
    
    @property
    def flags(self):
        return self._flags
    
    @flags.setter
    def flags(self, value):
        if self._flags != value:
            new_value = self.itemChange(self.ItemFlagsChange, value)
            self._flags = new_value
            self.itemChange(self.ItemFlagsHaveChanged, new_value)
    
    # Position property
    @property
    def position(self):
        return self._position
    
    @position.setter
    def position(self, value):
        if self._position != value:
            new_value = self.itemChange(self.ItemPositionChange, value)
            self._position = new_value
            self.itemChange(self.ItemPositionHasChanged, new_value)
    
    # Visibility property
    @property
    def visible(self):
        return self._visible
    
    @visible.setter
    def visible(self, value):
        if self._visible != value:
            new_value = self.itemChange(self.ItemVisibleChange, value)
            self._visible = new_value
            self.itemChange(self.ItemVisibleHasChanged, new_value)
    
    # Enabled property
    @property
    def enabled(self):
        return self._enabled
    
    @enabled.setter
    def enabled(self, value):
        if self._enabled != value:
            new_value = self.itemChange(self.ItemEnabledChange, value)
            self._enabled = new_value
            self.itemChange(self.ItemEnabledHasChanged, new_value)
    
    # Selected property
    @property
    def selected(self):
        return self._selected
    
    @selected.setter
    def selected(self, value):
        if self._selected != value:
            new_value = self.itemChange(self.ItemSelectedChange, value)
            self._selected = new_value
            self.itemChange(self.ItemSelectedHasChanged, new_value)
    
    # Transform property
    @property
    def transform(self):
        return self._transform
    
    @transform.setter
    def transform(self, value):
        if self._transform != value:
            new_value = self.itemChange(self.ItemTransformChange, value)
            self._transform = new_value
            self.itemChange(self.ItemTransformHasChanged, new_value)
    
    # Rotation property
    @property
    def rotation(self):
        return self._rotation
    
    @rotation.setter
    def rotation(self, value):
        if self._rotation != value:
            new_value = self.itemChange(self.ItemRotationChange, value)
            self._rotation = new_value
            self.itemChange(self.ItemRotationHasChanged, new_value)
    
    # Scale property
    @property
    def scale(self):
        return self._scale
    
    @scale.setter
    def scale(self, value):
        if self._scale != value:
            new_value = self.itemChange(self.ItemScaleChange, value)
            self._scale = new_value
            self.itemChange(self.ItemScaleHasChanged, new_value)
    
    # Transform origin point property
    @property
    def transformOriginPoint(self):
        return self._transform_origin
    
    @transformOriginPoint.setter
    def transformOriginPoint(self, value):
        if self._transform_origin != value:
            new_value = self.itemChange(self.ItemTransformOriginPointChange, value)
            self._transform_origin = new_value
            self.itemChange(self.ItemTransformOriginPointHasChanged, new_value)
    
    # Z-value property
    @property
    def zValue(self):
        return self._z_value
    
    @zValue.setter
    def zValue(self, value):
        if self._z_value != value:
            new_value = self.itemChange(self.ItemZValueChange, value)
            self._z_value = new_value
            self.itemChange(self.ItemZValueHasChanged, new_value)
    
    # Opacity property
    @property
    def opacity(self):
        return self._opacity
    
    @opacity.setter
    def opacity(self, value):
        if self._opacity != value:
            new_value = self.itemChange(self.ItemOpacityChange, value)
            self._opacity = new_value
            self.itemChange(self.ItemOpacityHasChanged, new_value)
    
    def setCacheMode(self, mode, logical_cache_size=None, device_cache_size=None):
        """
        Set the cache mode for this item.
        Optionally provide cache sizes for ItemCoordinateCache or DeviceCoordinateCache modes.
        """
        self._cache_mode = mode
        # Here you would typically store the cache sizes if needed
        self._logical_cache_size = logical_cache_size
        self._device_cache_size = device_cache_size
    
    def setFlag(self, flag, enabled=True):
        """Set or clear a flag for this item."""
        flag_int = flag.value if hasattr(flag, 'value') else flag
        if enabled:
            new_flags = self._flags | flag_int
        else:
            new_flags = self._flags & ~flag_int
        
        if self._flags != new_flags:
            self.flags = new_flags
    
    def setParent(self, parent):
        """
        Override setParent to handle parent change notifications
        """
        old_parent = self.parent()
        super().setParent(parent)
        if old_parent != parent:
            self.itemChange(self.ItemParentChange, parent)
            self.itemChange(self.ItemParentHasChanged, parent)
            
    def itemChange(self, change, value):
        """
        Handle item changes. Can be overridden in subclasses for custom behavior.
        Returns the new value after processing.
        """
        self.itemChanged.emit(change, value)
        return value




#%%%

import numpy as np
from dataclasses import dataclass
from typing import List, Dict, Tuple, Union, ClassVar
from pprint import pprint
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from collections import defaultdict

@dataclass(frozen=True, slots=True)
class AxisGeometry:
    __ax_to_index: ClassVar[Dict[str, int]] = {"x": 0, "y": 1, "z": 2}
    __index_to_ax: ClassVar[Dict[int, str]] = {0: "x", 1: "y", 2: "z"}
    
    @classmethod
    def getIndex(self, ax: str | int) -> int:
        return AxisGeometry.__ax_to_index.get(ax, None)
    
    @classmethod
    def getAxis(self, idx: int) -> str:
        return AxisGeometry.__index_to_ax.get(idx, None)
    
    @classmethod
    def axesIndex(cls):
        return list(AxisGeometry.__index_to_ax.keys())
    
    @classmethod
    def axesString(cls):
        return list(AxisGeometry.__ax_to_index.keys())



def verifyDirection(direction: str|int) -> int:
    if isinstance(direction, str):
        if direction not in AxisGeometry.axesIndex():
            raise ValueError("The direction is not one of x, y or z")
        else:
            direction = AxisGeometry.__ax_to_index(direction)            
    elif isinstance(direction, (int, float)) and not isinstance(direction, bool):
        if not direction in (0, 1, 2):
            raise ValueError("The direction is not one of axes indexes 0, 1, 2 (corresponds to: x, y, z)")            
    else:
        raise ValueError("The direction is not one of axes. Use one of 0, 1, 2 or x, y, z")
    return direction


def verifyAxisOrthoPoints(viewRange, axisVertexPoints):
    required_keys = {"direction", "coplanar", "perpendicular"}
    if set(axisVertexPoints) != required_keys:
        raise ValueError(f"axisOrthoPoints requires keys {required_keys}")
    
    value_to_keys = defaultdict(list)
    for key, val in axisVertexPoints.items():
        value_to_keys[val].append(key)
    
    value_to_rows = defaultdict(list)
    for i, row in enumerate(viewRange):
        for val in row:
            value_to_rows[val].append(i)
    
    overused_values = []
    for val, keys in value_to_keys.items():
        needed = len(keys)
        available = len(value_to_rows[val])
        if needed > available:
            overused_values.append((val, keys, needed, available))
    
    if overused_values:
        msgs = []
        for val, keys, needed, available in overused_values:
            keys_str = ", ".join(keys)
            msgs.append(f"value {val} used by [{keys_str}] requires {needed} distinct matches but only found {available}")
        raise ValueError("Invalid axisOrthoPoints mapping: " + " and ".join(msgs))


def verifyExtrumums(extremums):
    
    if len(extremums) != 0:
        raise ValueError(f"3 extremums are required, currently {len(extremums)}")
    
    if isinstance(extremums, dict):
        extremums_str = extremums.copy()
        extremums = 3*[None]
        for ax_str, ext in extremums_str.items():
            ax_idx = AxisGeometry.getIndex(ax_str)
            extremums[ax_idx] = ext
    
    if not all([ext in [0, 1] for ext in extremums]):
        raise ValueError("extremum values must be either 0 or 1")
    return extremums


def verifyCoplanarDirection(direction, coplanarDirection):
    if coplanarDirection is not None and direction == coplanarDirection:
        raise ValueError(f"coplanarDirection must be different to direction")

    if not coplanarDirection in AxisGeometry.axesIndex() + AxisGeometry.axesString():
        raise ValueError(f"coplanarDirection must be either one of (0, 1, 2) or ('x', 'y', 'z')")
    
    if isinstance(coplanarDirection, str):
        coplanarDirection = AxisGeometry.getIndex(coplanarDirection)
        
    return coplanarDirection


def verifyVertexPointsDirection(viewRange, direction, vertexPoints) -> int:
    diff_ortho_vertices = np.array(vertexPoints["coplanar"]) - np.array(vertexPoints["perpendicular"])
    
    if  np.sum(diff_ortho_vertices == 0) > 1:
        raise ValueError(f"Coplanar and perpendicular points: {vertexPoints} are not orthogonal to each other")
    
    directionFromVertices = diff_ortho_vertices[diff_ortho_vertices == 0].astype(int)
    
    if direction is not None and direction != directionFromVertices:
        raise ValueError(f"The given axis-direction {direction} is not the axis for the axisVertexPoints")
    
    if direction is None:
        direction = directionFromVertices
    
    for ax, point in vertexPoints.items():
        for idx, p in enumerate(point):
            lim = viewRange[idx]
            if not p in lim:
                raise ValueError(f"Vertex direction: {ax} with point {point} is not in viewRange {lim}")
    
    return direction
                
                


def fromViewBoxRange(viewRange: list[list[float]] | Dict[str, list[float]],
                     direction: str | int=None,
                     axisVertexPoints: Dict[str, float] = None,
                     extremums: Tuple[int, int, int]|Dict[str, int]=None, #    extremum values must be either 0 or 1, and orients the axis origin in world coordinates. {"x" : int, "y" : int ,..} or (int, int ,int)
                     coplanarDirection: str | int | None = None,
                     ) -> Tuple[Tuple[float, float, float],
                                Tuple[float, float, float],
                                Tuple[float, float, float]
                                ]:
    
    if direction is None and axisVertexPoints is None:
        raise ValueError(".fromViewBoxRange requires either axisVertexPoints or direction and extremum") 
    
    if direction is not None:
        direction = verifyDirection(direction)
    
    if axisVertexPoints is not None:
        direction = verifyVertexPointsDirection(axisVertexPoints)
        axisOrigin = axisOrthoPoints["direction"].copy()
        axisOrigin[direction] = [vr for vr in worldRange[direction] if vr != axisVertexPoints["direction"][direction]][0]
        
        coplanarDirection = np.array(axisVertexPoints["coplanar"]) - np.array(axisOrigin) 
    
    else:
        
        if coplanarDirection is None or extremums is None:
            raise ValueError(f"If axisVertexPoints is None, coplanarDirection and extremums needs to be provided")
        else:
            coplanarDirection = verifyCoplanarDirection(coplanarDirection)
            extremums = verifyExtrumums(extremums)
            perpendicularDirection = np.delete([0, 1, 2], [direction, coplanarDirection]).item()
            
            
            axisOrigin = [None, None, None]
            
            for idx, ext in enumerate(extremums):
                axisOrigin[idx] = worldRange[idx][ext]

            directionVertex = axisOrigin.copy()
            worldRangeIdx = (extremums[idx] - i) // 2 
            directionVertex[direction] = worldRange[direction][worldRangeIdx]

            coplanarVertex = axisOrigin.copy()
            worldRangeIdx = (extremums[coplanarDirection] - i) // 2 
            coplanarVertex[coplanarDirection] = worldRange[coplanarDirection][worldRangeIdx]
            
            perpendicularVertex = axisOrigin.copy()
            worldRangeIdx = (extremums[perpendicularDirection] - i) // 2 
            perpendicularVertex[perpendicularDirection] = worldRange[perpendicularDirection][worldRangeIdx]

            axisVertexPoints = {"direction" : directionVertex,
                                "coplanarDirection" : coplanarVertex,
                                "perpendicularDirection" : perpendicularVertex
                                }
            
            return axisVertexPoints    
        
        

    if axisVertexPoints is not None:
        
        verifyAxisOrthoPoints(viewRange, axisVertexPoints)
        
        
        directionOriginValue = viewRange["direction"][0] if axisVertexPoints["direction"] == viewRange[direction][1] else viewRange["direction"][1]
        
        directionOrigin = axisVertexPoints["direction"].copy()
        directionOrigin[direction] = directionOriginValue
        
        if axisOrigin is not None:
            if tuple(axisOrigin) != tuple(directionOrigin):
                raise ValueError('the origin from axisOrthoPoints does not match the axisOrigin. Either set axisOrigin==None or set axisOrthoPoints==None and provide coplanarOrientation and perpendicularOrientation')
        
        
        coplanarDirection = [idx for idx, value in enumerate(directionOrigin) if value - axisVertexPoints["coplanar"][idx] != 0][0]
        perpendicularDirection = [idx for idx, value in enumerate(directionOrigin) if value - axisVertexPoints["perpendicular"][idx] != 0][0]

        
        
            
            
        
        axisOrigin = axisVertexPoints["direction"].copy()
        axisOrigin["direction"] = axisOrigin[:]
        
        
        
    
    if isinstance(viewRange, dict):
        viewRange = [viewRange['x'], viewRange['y'], viewRange['z']]
        
    if isinstance(coplanarDirection, str):
        if coplanarDirection not in AxisGeometry.axesIndex():
            raise ValueError("The coplanarDirection is not one of x, y or z")
        else:
            coplanarDirection = AxisGeometry.__ax_to_index(coplanarDirection)
    
    if isinstance(perpendicularDirection, str):
        if perpendicularDirection not in AxisGeometry.axesIndex():
            raise ValueError("The perpendicularDirection is not one of x, y or z")
        else:
            perpendicularDirection = AxisGeometry.__ax_to_index(perpendicularDirection)
    
    if coplanarDirection is None:
        coplanarDirection = [i for i in AxisGeometry.axesIndex() if i != direction and i != perpendicularDirection][0]
    
    elif perpendicularDirection is None:
        perpendicularDirection = [i for i in AxisGeometry.axesIndex() if i != direction and i != coplanarDirection][0]
    
    
    if all((coplanarDirection is None, perpendicularDirection is None)):
        raise ValueError("A coplanarDirection or perpendicularDirection is required")
    else:
        if len((direction, coplanarDirection, perpendicularDirection)) != len (set((direction, coplanarDirection, perpendicularDirection))):
            raise ValueError(f"direction, coplanarDirection and perpendicularDirection must all be different, given {(direction, coplanarDirection, perpendicularDirection)}")

    if axisOrigin is None:
        if all((coplanarDirection is None, perpendicularDirection is None)):
            raise ValueError("A coplanarDirection or perpendicularDirection is required")

        if any((coplanarOrientation is None, perpendicularOrientation is None)):
            raise ValueError("A coplanarOrientation and perpendicularOrientation must be provided when axisOrigin is not provided")

        
        perpendicularRange = viewRange[perpendicularDirection][1] - viewRange[perpendicularDirection][0]
    
    
        if perpendicularRange / abs(perpendicularRange) != perpendicularOrientation:
            ValueError("A perpendicularOrientation does not match the viewRange dimensions")
        
        coplanarRange = viewRange[coplanarDirection][1] - viewRange[coplanarDirection][0]
        
        if coplanarRange / abs(coplanarRange) != coplanarOrientation:
            ValueError("A coplanarOrientation does not match the viewRange dimensions")

        

        axisOrigin = [0, 0, 0]
        axisOrigin[direction] = viewRange[direction][0]
        
        perpViewRange = viewRange[perpendicularDirection]
        
        idx = (perpendicularOrientation - 1) // 2
        perpendicularAtOrigin = perpViewRange[idx]
        
        axisOrigin[perpendicularDirection] = perpendicularAtOrigin
        
        coplanarViewRange = viewRange[coplanarDirection]
        
        idx = (coplanarOrientation - 1) // 2
        coplanarAtOrigin = coplanarViewRange[idx]
        axisOrigin[coplanarDirection] = coplanarAtOrigin
    else:
        
        for idx, coord_value in enumerate(axisOrigin):
            if coord_value not in viewRange[idx]:
                raise ValueError(f"axisOrigin component {idx} ({coord_value}) not in viewRange {viewRange[idx]}")

    coplanarEdge = axisOrigin.copy()
    idx = (coplanarOrientation + 1) // 2        
    coplanarEdge[coplanarDirection] = viewRange[coplanarDirection][idx]
    
    idx = (perpendicularOrientation + 1) // 2
    perpendicularEdge = axisOrigin.copy()
    perpendicularEdge[perpendicularDirection] = viewRange[perpendicularDirection][idx]
    
    axisEdge = axisOrigin.copy()
    
    directionMaxValue = viewRange[direction][0] if viewRange[direction][1] == axisOrigin[direction] else viewRange[direction][1]
    
    axisEdge[direction] = directionMaxValue

            
    points = {"direction_0" : np.round(tuple(axisOrigin), 1).tolist(),
              "dir" : np.round(axisEdge, 1).tolist(),
              "coplanar" : np.round(tuple(coplanarEdge), 1).tolist(),
              "perp" : np.round(tuple(perpendicularEdge), 1).tolist(),
              
              }
    
    
    return points


def plot_points(points):

    origin = np.array(points["direction_0"])

    # Setup plot
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    # Plot origin
    ax.scatter(*origin, color='k')#   , label="direction_0 (origin)")

    # Define colors for each point
    colors = {
        "dir": "r",
        "coplanar": "g",
        "perp": "b"
    }

    # Plot other points and lines from origin
    for name, point in points.items():
        point = np.array(point)
        if name != "direction_0":
            ax.scatter(*point, color=colors[name], label=name)
            ax.plot([origin[0], point[0]], [origin[1], point[1]], [origin[2], point[2]], color=colors[name])

    # Axes labels and legend
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    ax.legend()

    #ax.set_box_aspect([1, 1, 1])  # Optional: Equal aspect ratio

    plt.tight_layout()
    plt.show()


worldRange=[[-0.2, 1.5],
            [-0.7, 1.],
            [-0.1,1.5]
            ]



direction = 1
coplanarDirection = 0
coplanarOrientation = 1


perpendicularDirection = 2
perpendicularOrientation = -1
axisOrthoPoints = {
    "direction": 1.,
    "coplanar": 1.0,
    "perpendicular": 1.  # <-- this is a duplicate
}

points = fromViewBoxRange(worldRange,
                          direction, 
                      #    axisOrigin=axisOrigin,
                          coplanarDirection=coplanarDirection,
                          coplanarOrientation=coplanarOrientation,
                          perpendicularDirection=perpendicularDirection,
                          perpendicularOrientation=perpendicularOrientation
                          )


plot_points(points)



pprint(worldRange, width=15)
print(f"\norientation:\ncoplanar: {coplanarOrientation}\nperp: {perpendicularOrientation}")

pprint(points)


# %%



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from instruments.utils import InstrumentManager
    
from PySide6 import QtCore, QtWidgets
from . import main_view
from . import settings
from .. import abstract_classes
from ...quantities import engines
from ...data_classes import builders as builders_data_classes
from ..abstract_classes import ABCInterface, ABC


QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)


class Interface(abstract_classes.ABCInterface):
    aboutToCloseSig = QtCore.Signal(QtWidgets.QMainWindow)
    def __init__(self,
                 instrument_manager: InstrumentManager=None,
                 surface_calibration_config: Dict[str, Dict[str, Any]]={},
                 tick_engine_manager=None,
                 vol_vect_container=None,
                 **configs
                 ):
        super().__init__()
        self.all_price_types = instrument_manager.config["options"]["price_types"]
        
        self.splot_flag=True
        self.subplots_flag=True
        self.on_view=False
        self.plot_interaction_buffer=[]
        
        if tick_engine_manager is None:
            self.tick_engine_manager = engines.TickEngineManager("Strike",
                                                                 "Date",
                                                                 "Implied Volatility"
                                                                 )
        else:
            self.tick_engine_manager=tick_engine_manager

        volatility_data_container = builders_data_classes.create_volatility_data_from_vol_vect(vol_vect_container)
        
        self.main_view: ABCInterface = main_view.MainView(instrument_manager,
                                                        engine_container=surface_calibration_config["engine_container"],
                                                        tick_engine_manager=self.tick_engine_manager,
                                                        spot_objects=instrument_manager.spot_instrument_container.get_objects(),
                                                        volatility_data_container=volatility_data_container,
                                                        **configs
                                                        )  
        #self.main_view.addIntenseInteraction("settings")
        
        self.setCentralWidget(self.main_view)
        settings = self._initSettings()
        
        self.addSettings(settings)
        
        settings.toggle_vol_src(self.all_price_types[0])
                    
        self.showMaximized()
    
        
    def addSettings(self, settings):
        if not settings is None:
            self.addToolBar(settings)
    
    
    def get_calibration_slot(self) -> Callable:
        return self.main_view.update_view

    def _initSettings(self) -> settings.Settings:
        settings_widget = settings.Settings(self)
     
        settings_widget.processIntenseInteraction=self.main_view.processInteractionState
        
        switch_axis_slots = [self.tick_engine_manager.change_function]
        switch_axis_slots = switch_axis_slots + [vola_data.vol_vector.metric_engine.change_function for vola_data in self.main_view.volatility_data_container.values()]
        switch_axis_slots = switch_axis_slots + [self.main_view.switch_axis]      
        
        settings_widget.add_window_menu(self.main_view.switch_axis_units)

        settings_widget.create_combobox_menu("Moneyness",
                                             ["Strike", "Delta", "Moneyness", "Log-Moneyness", "Standardised-Moneyness"],
                                             switch_axis_slots,
                                             "x"
                                             )

        settings_widget.create_combobox_menu("Tenor",
                                             ["Date", "Years"],
                                             switch_axis_slots,
                                             "y"
                                             )

        settings_widget.create_combobox_menu("Volatility",
                                             ["Implied Volatility", "Variance", "Total Volatility", "Total Variance"],
                                             switch_axis_slots,
                                             "z"
                                             )

        settings_widget.create_combobox_menu("Toggle Crosshairs",
                                             ["On", "Off"],
                                             self.main_view.pyvol_gl_widget.toggle_crosshairs,
                                             )
        
        settings_widget.create_vol_src_menu(self.all_price_types,
                                            [self.main_view.toggle_price_type]
                                            )

        settings_widget.create_toggle_buttons("Toggle 3D Assets",
                                              ["surface", "scatter"], 
                                              [self.main_view.toggle_3D_objects],
                                              )
        return settings_widget
    
    



from PySide6.QtCore import QObject, Signal
from PySide6.QtWidgets import QGraphicsItem
from PySide6 import QtWidgets


class EnumContainer:
    
    ItemCoordinateCache = QGraphicsItem.CacheMode.ItemCoordinateCache
    DeviceCoordinateCache = QGraphicsItem.CacheMode.DeviceCoordinateCache
    
    # GraphicsItemChange enums
    ItemEnabledChange = QGraphicsItem.GraphicsItemChange.ItemEnabledChange
    ItemEnabledHasChanged = QGraphicsItem.GraphicsItemChange.ItemEnabledHasChanged
    ItemPositionChange = QGraphicsItem.GraphicsItemChange.ItemPositionChange
    ItemPositionHasChanged = QGraphicsItem.GraphicsItemChange.ItemPositionHasChanged
    ItemTransformChange = QGraphicsItem.GraphicsItemChange.ItemTransformChange
    ItemTransformHasChanged = QGraphicsItem.GraphicsItemChange.ItemTransformHasChanged
    ItemRotationChange = QGraphicsItem.GraphicsItemChange.ItemRotationChange
    ItemRotationHasChanged = QGraphicsItem.GraphicsItemChange.ItemRotationHasChanged
    ItemScaleChange = QGraphicsItem.GraphicsItemChange.ItemScaleChange
    ItemScaleHasChanged = QGraphicsItem.GraphicsItemChange.ItemScaleHasChanged
    ItemTransformOriginPointChange = QGraphicsItem.GraphicsItemChange.ItemTransformOriginPointChange
    ItemTransformOriginPointHasChanged = QGraphicsItem.GraphicsItemChange.ItemTransformOriginPointHasChanged
    ItemSelectedChange = QGraphicsItem.GraphicsItemChange.ItemSelectedChange
    ItemSelectedHasChanged = QGraphicsItem.GraphicsItemChange.ItemSelectedHasChanged
    ItemVisibleChange = QGraphicsItem.GraphicsItemChange.ItemVisibleChange
    ItemVisibleHasChanged = QGraphicsItem.GraphicsItemChange.ItemVisibleHasChanged
    ItemParentChange = QGraphicsItem.GraphicsItemChange.ItemParentChange
    ItemParentHasChanged = QGraphicsItem.GraphicsItemChange.ItemParentHasChanged
    ItemChildAddedChange = QGraphicsItem.GraphicsItemChange.ItemChildAddedChange
    ItemChildRemovedChange = QGraphicsItem.GraphicsItemChange.ItemChildRemovedChange
    ItemSceneChange = QGraphicsItem.GraphicsItemChange.ItemSceneChange
    ItemSceneHasChanged = QGraphicsItem.GraphicsItemChange.ItemSceneHasChanged
    ItemCursorChange = QGraphicsItem.GraphicsItemChange.ItemCursorChange
    ItemCursorHasChanged = QGraphicsItem.GraphicsItemChange.ItemCursorHasChanged
    ItemToolTipChange = QGraphicsItem.GraphicsItemChange.ItemToolTipChange
    ItemToolTipHasChanged = QGraphicsItem.GraphicsItemChange.ItemToolTipHasChanged
    ItemFlagsChange = QGraphicsItem.GraphicsItemChange.ItemFlagsChange
    ItemFlagsHaveChanged = QGraphicsItem.GraphicsItemChange.ItemFlagsHaveChanged
    ItemOpacityChange = QGraphicsItem.GraphicsItemChange.ItemOpacityChange
    ItemOpacityHasChanged = QGraphicsItem.GraphicsItemChange.ItemOpacityHasChanged
    ItemScenePositionHasChanged = QGraphicsItem.GraphicsItemChange.ItemScenePositionHasChanged

    ItemIsMovable = QGraphicsItem.GraphicsItemFlag.ItemIsMovable.value
    ItemIsSelectable = QGraphicsItem.GraphicsItemFlag.ItemIsSelectable.value
    ItemIsFocusable = QGraphicsItem.GraphicsItemFlag.ItemIsFocusable.value
    ItemClipsToShape = QGraphicsItem.GraphicsItemFlag.ItemClipsToShape.value
    ItemClipsChildrenToShape = QGraphicsItem.GraphicsItemFlag.ItemClipsChildrenToShape.value
    ItemIgnoresTransformations = QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations.value
    ItemIgnoresParentOpacity = QGraphicsItem.GraphicsItemFlag.ItemIgnoresParentOpacity.value
    ItemDoesntPropagateOpacityToChildren = QGraphicsItem.GraphicsItemFlag.ItemDoesntPropagateOpacityToChildren.value
    ItemStacksBehindParent = QGraphicsItem.GraphicsItemFlag.ItemStacksBehindParent.value
    ItemUsesExtendedStyleOption = QGraphicsItem.GraphicsItemFlag.ItemUsesExtendedStyleOption.value
    ItemHasNoContents = QGraphicsItem.GraphicsItemFlag.ItemHasNoContents.value
    ItemSendsGeometryChanges = QGraphicsItem.GraphicsItemFlag.ItemSendsGeometryChanges.value
    ItemAcceptsInputMethod = QGraphicsItem.GraphicsItemFlag.ItemAcceptsInputMethod.value
    ItemIsPanel = QGraphicsItem.GraphicsItemFlag.ItemIsPanel.value
    ItemSendsScenePositionChanges = QGraphicsItem.GraphicsItemFlag.ItemSendsScenePositionChanges.value
    ItemContainsChildrenInShape = QGraphicsItem.GraphicsItemFlag.ItemContainsChildrenInShape.value
    
class CustomGraphicsItem(QGraphicsItem):
    def itemChange(self, change, value):
        print(f"change: {change}")
        print(f"value: {value}")

        ret = super().itemChange(change, value)
        print(f"ret: {ret}")
        return 

class CustomGraphicsItem2(QGraphicsItem):
    """
    A custom QObject class that emulates QGraphicsItem's cache modes,
    item change enums, and flags functionality.
    """
    
    # Cache modes
    itemChanged = Signal(QGraphicsItem.GraphicsItemChange, object)  # change, value
    
    def __init__(self, parent=None):
        super().__init__(parent)
        self._cache_mode = self.NoCache
        self._flags = 0
        self._position = (0, 0)
        self._visible = True
        self._enabled = True
        self._selected = False
        self._transform = None
        self._rotation = 0.0
        self._scale = 1.0
        self._transform_origin = (0, 0)
        self._z_value = 0.0
        self._opacity = 1.0
        
    def itemChange(self, change, value):
        print(f"change: {change}")
        print(f"value: {value}")

        ret = QGraphicsItem.GraphicsItemChange(change, value)
        print(f"ret: {ret}")
        if change == CustomGraphicsItem.ItemPositionChange:
            print(f"Item changed: {change}, New value: {value}")
        """
        Handle item changes. Can be overridden in subclasses for custom behavior.
        Returns the new value after processing.
        """
        #self.itemChanged.emit(change, value)
        return value
    
    # Cache mode property
    @property
    def cacheMode(self):
        return self._cache_mode
    
    @cacheMode.setter
    def cacheMode(self, value):
        if self._cache_mode != value:
            self._cache_mode = value
    
    @property
    def flags(self):
        return self._flags
    
    @flags.setter
    def flags(self, value):
        if self._flags != value:
            new_value = self.itemChange(self.ItemFlagsChange, value)
            self._flags = new_value
            self.itemChange(self.ItemFlagsHaveChanged, new_value)
    
    # Position property
    @property
    def position(self):
        return self._position
    
    @position.setter
    def position(self, value):
        if self._position != value:
            new_value = self.itemChange(self.ItemPositionChange, value)
            self._position = new_value
            self.itemChange(self.ItemPositionHasChanged, new_value)
    
    # Visibility property
    @property
    def visible(self):
        return self._visible
    
    @visible.setter
    def visible(self, value):
        if self._visible != value:
            new_value = self.itemChange(self.ItemVisibleChange, value)
            self._visible = new_value
            self.itemChange(self.ItemVisibleHasChanged, new_value)
    
    # Enabled property
    @property
    def enabled(self):
        return self._enabled
    
    @enabled.setter
    def enabled(self, value):
        if self._enabled != value:
            new_value = self.itemChange(self.ItemEnabledChange, value)
            self._enabled = new_value
            self.itemChange(self.ItemEnabledHasChanged, new_value)
    
    # Selected property
    @property
    def selected(self):
        return self._selected
    
    @selected.setter
    def selected(self, value):
        if self._selected != value:
            new_value = self.itemChange(self.ItemSelectedChange, value)
            self._selected = new_value
            self.itemChange(self.ItemSelectedHasChanged, new_value)
    
    # Transform property
    @property
    def transform(self):
        return self._transform
    
    @transform.setter
    def transform(self, value):
        if self._transform != value:
            new_value = self.itemChange(self.ItemTransformChange, value)
            self._transform = new_value
            self.itemChange(self.ItemTransformHasChanged, new_value)
    
    # Rotation property
    @property
    def rotation(self):
        return self._rotation
    
    @rotation.setter
    def rotation(self, value):
        if self._rotation != value:
            new_value = self.itemChange(self.ItemRotationChange, value)
            self._rotation = new_value
            self.itemChange(self.ItemRotationHasChanged, new_value)
    
    # Scale property
    @property
    def scale(self):
        return self._scale
    
    @scale.setter
    def scale(self, value):
        if self._scale != value:
            new_value = self.itemChange(self.ItemScaleChange, value)
            self._scale = new_value
            self.itemChange(self.ItemScaleHasChanged, new_value)
    
    # Transform origin point property
    @property
    def transformOriginPoint(self):
        return self._transform_origin
    
    @transformOriginPoint.setter
    def transformOriginPoint(self, value):
        if self._transform_origin != value:
            new_value = self.itemChange(self.ItemTransformOriginPointChange, value)
            self._transform_origin = new_value
            self.itemChange(self.ItemTransformOriginPointHasChanged, new_value)
    
    # Opacity property
    @property
    def opacity(self):
        return self._opacity
    
    @opacity.setter
    def opacity(self, value):
        if self._opacity != value:
            new_value = self.itemChange(self.ItemOpacityChange, value)
            self._opacity = new_value
            self.itemChange(self.ItemOpacityHasChanged, new_value)
    
    def setCacheMode(self, mode, logical_cache_size=None, device_cache_size=None):
        """
        Set the cache mode for this item.
        Optionally provide cache sizes for ItemCoordinateCache or DeviceCoordinateCache modes.
        """
        self._cache_mode = mode
        # Here you would typically store the cache sizes if needed
        self._logical_cache_size = logical_cache_size
        self._device_cache_size = device_cache_size
    
    def setFlag(self, flag, enabled=True):
        """Set or clear a flag for this item."""
        flag_int = flag.value if hasattr(flag, 'value') else flag
        if enabled:
            new_flags = self._flags | flag_int
        else:
            new_flags = self._flags & ~flag_int
        
        if self._flags != new_flags:
            self.flags = new_flags
    
    def setParent(self, parent):
        """
        Override setParent to handle parent change notifications
        """
        old_parent = self.parent()
        super().setParent(parent)
        if old_parent != parent:
            self.itemChange(self.ItemParentChange, parent)
            self.itemChange(self.ItemParentHasChanged, parent)
            
def handle_item_change(change, value):
    # Map change enums to names for demonstration
    change_names = {
        CustomGraphicsItem.ItemPositionChange: "PositionChange",
        CustomGraphicsItem.ItemPositionHasChanged: "PositionHasChanged",
        CustomGraphicsItem.ItemVisibleChange: "VisibleChange",
        CustomGraphicsItem.ItemSelectedChange: "SelectedChange",
        CustomGraphicsItem.ItemFlagsChange: "FlagsChange",
        # Add other change types as needed
    }
    print(f"Item changed: {change_names.get(change, change)}, New value: {value}")


app = QtWidgets.QApplication([])

item = CustomGraphicsItem()

# Connect to the itemChanged signal
#item.itemChanged.connect(handle_item_change)

# Set some flags
item.setFlag(CustomGraphicsItem.ItemIsMovable)
item.setFlag(CustomGraphicsItem.ItemIsSelectable)
item.setFlag(CustomGraphicsItem.ItemSendsGeometryChanges)

# Make changes that will trigger notifications
item.position = (100, 100)  # Will emit PositionChange and PositionHasChanged
item.visible = False        # Will emit VisibleChange
item.selected = True        # Will emit SelectedChange
item.flags = CustomGraphicsItem.ItemIsMovable | CustomGraphicsItem.ItemIsSelectable  # Will emit FlagsChange

# Set cache mode
item.setCacheMode(CustomGraphicsItem.DeviceCoordinateCache)


#app.exec()



from PySide6 import QtWidgets, QtGui
import numpy as np, sys
from pyqtgraph.opengl import GLLinePlotItem, GLViewWidget, GLTextItem
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
from pyqtgraph import Vector

class BlockUpdateGLLinePlot(GLLinePlotItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)   

class NestedLineItem(GLLinePlotItem):
    def __init__(self, *args, **kwargs):
        self._otherNests=[]
        self._lineItems=[]
        self.blockUpdates=False
        self.blockPaint=False
        super().__init__(*args, **kwargs)   

    def addNest(self, nestedLineItem):
        self._otherNests.append(nestedLineItem)

    def addLineItem(self, lineItem):
        self._lineItems.append(lineItem)

    def _extract(self, combined_pos):
        T = (self.parentItem().transform() * self.transform()).inverted()[0]
        for nestedLineItem in self._otherNests:
            for lineItem in nestedLineItem._lineItems:
                pos = [T.map(QtGui.QVector3D(p[0], p[1], p[2])) for p in lineItem.pos]
                
                pos = [nestedLineItem.parentItem().transform().map(QtGui.QVector3D(p.x(), p.y(), p.z())) for p in pos]
                pos = [nestedLineItem.transform().map(QtGui.QVector3D(p.x(), p.y(), p.z())) for p in pos]
                pos = np.array([[p.x(), p.y(), p.z()] for p in pos])
                
                pos = self._mapLineItem(pos, lineItem.transform())
                combined_pos = pos if combined_pos is None else np.vstack((combined_pos, pos))
                
            self.last_child2pos = pos[-1]
        return combined_pos

    def paint(self):
        if not self.blockPaint:
            combined_pos = None
            for lineItem in self._lineItems:
                pos = self._mapLineItem(lineItem.pos.copy(), lineItem.transform())
                combined_pos = pos if combined_pos is None else np.vstack((combined_pos, pos))
            
            combined_pos = self._extract(combined_pos)

            self.blockUpdates=True
            self.setData(pos=combined_pos)
            self.blockUpdates=True
            super().paint()

    def _mapLineItem(self, pos, tr):        
        if not tr.isIdentity():
            points_h = np.hstack([pos, np.ones((pos.shape[0], 1))])
            tr_matrix = np.array(tr.data()).reshape(4, 4).T
            transformed_h = points_h @ tr_matrix.T
            pos = transformed_h[:, :3] / transformed_h[:, 3:4]  
        return pos
    
    def update(self):
        if not self.blockUpdates:
            super().update()    
    
class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, app=None, **kwargs):
        super().__init__(*args, **kwargs)   
        self.gl_widget = GLViewWidget()
        self.setCentralWidget(self.gl_widget)
        
        self.gP1, self.gP2, self.gP22 = GLGraphicsItem(), GLGraphicsItem(), GLGraphicsItem()        
    #    self.gP1.translate(1, 1, 1), self.gP2.translate(1, 3, -1), self.gP22.translate(1, 3, -1)
       # self.gP2.scale(-1, 1, -1)
        #self.gP22.scale(-1, 1, -1)


        self.nestPar1, self.nestPar2, self.nestPar22 = NestedLineItem(color="red", mode="lines"), NestedLineItem(color="yellow", mode="lines"), NestedLineItem(color="orange", mode="lines")
        self.nestPar1.translate(0.2, 1, 1), self.nestPar2.translate(-0.2, 5, 0.2), self.nestPar22.translate(-0.2, 5, 0.2)
        
        self.nestPar2.scale(-1, 1, -1), self.nestPar22.scale(-1, 1, -1)
        
        self.nestPar1.setParentItem(self.gP1), self.nestPar2.setParentItem(self.gP2), self.nestPar22.setParentItem(self.gP22)
        
        lineLength = 1
        
        for z in np.linspace(0, 5, 6):
            pos1 = [[0, 0, z], [lineLength, 0, z]]
            pos2 = [[0, 1, z ], [lineLength, 1, z]]  
            pos22 = [[0, 1, z+0.1], [lineLength, 1,  z+0.1]] # small offset
            l1, l2, l22 = GLLinePlotItem(pos=pos1, mode="lines"), GLLinePlotItem(pos=pos2, mode="lines"), GLLinePlotItem(pos=pos22, mode="lines")
            
            translation = -0.5, -1, 0.5
          #  l2.scale(-1, 1, -1), l22.scale(-1, 1, -1)
         #   l1.translate(*translation, True), l2.translate(*translation, True), l22.translate(*translation, True)
            self.nestPar1.addLineItem(l1), self.nestPar2.addLineItem(l2), self.nestPar22.addLineItem(l22)
        
        self.nestPar2.blockPaint=True
        self.nestPar1.addNest(self.nestPar2)
        
        self.gl_widget.addItem(self.gP1)
        self.gl_widget.addItem(self.gP22)
        
        self.showMaximized()
        self.nestPar1.paint()
        
        text2 = GLTextItem(pos=self.nestPar1.last_child2pos, text="mapping of child2 (incorrect)", font=QtGui.QFont("Arial", 6))
        text2.setTransform(self.nestPar1.transform())
        text2.setParentItem(self.gP1)
        
        text22 = GLTextItem(pos=pos22[-1], text="Child2 should be here", font=QtGui.QFont("Arial", 6))
        text22.translate(*translation, True)
        text22.setParentItem(self.gP22)
        
        
if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    win = Window()
    sys.exit(app.exec())


import sys
import time
from PySide6.QtWidgets import QApplication, QVBoxLayout, QWidget, QPushButton, QLabel
import pyqtgraph as pg
import pyqtgraph.opengl as gl
import numpy as np


class GLTextItemPerformanceTest(QWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("GLTextItem Performance Test")
        self.setGeometry(100, 100, 1000, 800)
        
        # Create OpenGL view
        self.gl_view = gl.GLViewWidget()
        self.gl_view.setCameraPosition(distance=50)
        
        # Add grid for reference
        grid = gl.GLGridItem()
        self.gl_view.addItem(grid)
        
        # Create text item
        self.text_item = gl.GLTextItem(text="Initial Text", pos=(0, 0, 0))
        self.gl_view.addItem(self.text_item)
        
        # Create UI controls
        self.btn_run_tests = QPushButton("Run Performance Tests")
        self.result_label = QLabel("Results will appear here")
        
        # Layout
        layout = QVBoxLayout()
        layout.addWidget(self.gl_view, 1)
        layout.addWidget(self.btn_run_tests)
        layout.addWidget(self.result_label)
        self.setLayout(layout)
        
        # Connect signals
        self.btn_run_tests.clicked.connect(self.run_all_tests)
        
        # Test parameters
        self.iterations = 10000
        self.texts = [f"Text {i}" for i in range(self.iterations)]
        self.positions = [np.array([np.sin(i*0.1)*10, np.cos(i*0.1)*10, 0]) 
                          for i in range(self.iterations)]
    
    def run_all_tests(self):
        """Run both tests and compare results"""
        self.btn_run_tests.setEnabled(False)
        self.result_label.setText("Running tests...")
        QApplication.processEvents()  # Update UI immediately
        
        # Run text+position test
        both_time = self.test_both_updates()
        
        # Run text-only test
        text_time = self.test_text_only_update()
        
        # Compare results
        self.show_comparison(both_time, text_time)
        self.btn_run_tests.setEnabled(True)
    
    def test_both_updates(self):
        """Test updating both text and position in one setData call"""
        start_time = time.perf_counter()
        
        for i in range(self.iterations):
            self.text_item.setData(text=self.texts[i], pos=self.positions[i])
            if i % 100 == 0:  # Update UI occasionally for responsiveness
                QApplication.processEvents()
                
        return time.perf_counter() - start_time
    
    def test_text_only_update(self):
        """Test updating only text property"""
        start_time = time.perf_counter()
        
        for i in range(self.iterations):
            self.text_item.setData(text=self.texts[i])
            if i % 100 == 0:  # Update UI occasionally for responsiveness
                QApplication.processEvents()
                
        return time.perf_counter() - start_time
    
    def show_comparison(self, both_time, text_time):
        """Display comparison of both test results"""
        both_avg = both_time / self.iterations * 1000  # ms
        text_avg = text_time / self.iterations * 1000  # ms
        speed_diff = both_avg - text_avg
        speed_ratio = both_avg / text_avg if text_avg != 0 else 0
        
        results = (
            f"=== Performance Comparison ===\n"
            f"Test iterations: {self.iterations}\n\n"
            f"BOTH text+position updates:\n"
            f"  Total time: {both_time:.4f}s\n"
            f"  Average: {both_avg:.6f} ms/update\n\n"
            f"ONLY text updates:\n"
            f"  Total time: {text_time:.4f}s\n"
            f"  Average: {text_avg:.6f} ms/update\n\n"
            f"Comparison:\n"
            f"  Text-only is {speed_diff:.3f} ms/update faster\n"
            f"  Text-only is {speed_ratio:.1f}x faster"
        )
        
        self.result_label.setText(results)
        print(results)

if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = GLTextItemPerformanceTest()
    window.show()
    sys.exit(app.exec())



"""
This file contains the non-constant variable requirements for specific quantities that are not quoted.
E.g. For calculating implied volatility we will need a price source, for calculating delta we will need an ivol source (which will then need a price source...).
We wouldn't need an expiry source since we already know that and we wouldn't need an underlying price source since that's already quoted. 

I have chosen this manner because otherwise we would be computing every quantity at runtime, which is obviously sub-optimal. 

"""

"ivol"  = "px_type"
"delta" = "ivol"
"gamma" = "ivol"
"vega" = "ivol"
"rho" = "ivol"
"theta" = "ivol"
"standardised_moneyness" = "ivol"


from PySide6 import QtWidgets


class Settings(QtWidgets.QWidget):
    def __init__(self, widget_data_display, tick_engine_manager=None, metric_axis_engine=None,  spot_qlabel=None, parent=None):
        super().__init__(parent)
        self.widget_data_display=widget_data_display
        self.setLayout(QtWidgets.QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0 ,0)
        self.combobox=None

        action_handlers = []
        self.create_moneyness_combobox(["Strike", "Moneyness (%)", "Log-Moneyness", "Standardised-Moneyness", "Delta"], "x", action_handlers)
        self.create_price_type_combobox(["bid", "ask", "mid"], [])
        
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        
        if spot_qlabel is not None:
            spot_qlabel.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
            self.layout().addWidget(spot_qlabel)
            
    def create_col_filter(self):
        self.line_edit = QtWidgets.QLineEdit()
        self.line_edit.setText("10")
        self.line_edit.setStyleSheet("""
                                    QLineEdit {
                                        background-color: white;
                                        color: black;
                                    }
                                    QLineEdit:focus {
                                        background-color: #ffffcc;  
                                        border: 2px solid #3366ff;
                                    }
                                    """)
        self.line_edit.editingFinished.connect(self.handle_line_edit)
        self.layout().addWidget(self.line_edit)
    
    def create_moneyness_combobox(self, options, axis_direction, action_handlers):
        combobox = QtWidgets.QComboBox()
        combobox.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        combobox.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        combobox.blockSignals(True)
        combobox.addItems(options)
        
        for action_handler in action_handlers:
            combobox.currentTextChanged.connect(lambda text, ax=axis_direction, handler=action_handler: handler(ax, text))
        
        self.layout().addWidget(combobox)
        combobox.blockSignals(False)
        self.combobox=combobox

    def create_price_type_combobox(self, options, action_handlers):
        combobox = QtWidgets.QComboBox()
        combobox.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        combobox.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        combobox.blockSignals(True)
        combobox.addItems(options)
        
        for action_handler in action_handlers:
            combobox.currentTextChanged.connect(lambda text, handler=action_handler:  handler(text))
        
        self.layout().addWidget(combobox)
        combobox.blockSignals(False)
        self.combobox=combobox

        
    
    def handle_line_edit(self):
        text = self.line_edit.text()
        try:
            value = int(text)
        except ValueError:
            self.line_edit.setText("10")
            value = 10

        if self.widget_data_display:
            self.widget_data_display.handle_line_edit_value(value)
        else:
            print(f"Line edit value: {value}")




def legend_colour_getter(expiry_list):
    
    list_of_colours = ["white", "red","blue",  "green", "yellow", "cyan", "magenta",
    "orange", "purple", "pink", "brown",
    
    "lightgray", "darkgray", "dimgray",
    "lightred", "darkred",
    "lightgreen", "darkgreen", "lime", "limegreen",
    "lightblue", "darkblue", "navy",
    "lightcyan", "darkcyan", "teal",
    "lightmagenta", "darkmagenta", "fuchsia",
    "lightyellow", "darkyellow", "gold", "goldenrod", "khaki",
    "lightpink", "darkpink", "hotpink",
    "lightsalmon", "darksalmon", "coral",
    "lightgoldenrod", "darkgoldenrod",]
    
    colourmap = {}
    for expiry, colour in zip(expiry_list, list_of_colours):
         colourmap[expiry] = colour
    return colourmap


import sys
import numpy as np
from PySide6.QtWidgets import QApplication, QMainWindow, QVBoxLayout, QWidget, QLabel
from PySide6.QtCore import Qt
import pyqtgraph as pg
import pyqtgraph.opengl as gl
from pyqtgraph.Qt import QtCore
from OpenGL import GL, GLU

def map_2D_coords_to_3D(widget, x: float, y: float):
    widget_width = widget.width()
    widget_height = widget.height()
    device_pixel_ratio = widget.window().screen().devicePixelRatio()

    ndc_x = x / widget_width
    ndc_y = y / widget_height

    viewport = GL.glGetIntegerv(GL.GL_VIEWPORT)
    
    _, _, viewport_width, viewport_height = viewport

    mouse_x_physical = ndc_x * viewport_width
    mouse_y_physical = ndc_y * viewport_height
    mouse_y_physical = viewport_height - mouse_y_physical 
    
    depth = GL.glReadPixels(int(mouse_x_physical), int(mouse_y_physical), 1, 1, GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT)[0][0]

    modelview = np.array(widget.viewMatrix().data()).reshape(4, 4)
    projection = np.array(widget.projectionMatrix(viewport, (0, 0, device_pixel_ratio * widget_width, device_pixel_ratio* widget_height)).data()).reshape(4, 4)
    
    world_x, world_y, world_z = GLU.gluUnProject(mouse_x_physical, mouse_y_physical, depth, modelview, projection, viewport)
    
    return pg.Vector(world_x, world_y, world_z)



class MeshPointIntersectionDemo(QMainWindow):
    def __init__(self):
        super().__init__()
        
        # Set up the main window
        self.setWindowTitle("Mesh-Point Intersection Demo")
        self.setGeometry(100, 100, 1000, 800)
        
        # Create central widget and layout
        central_widget = QWidget()
        self.setCentralWidget(central_widget)
        layout = QVBoxLayout(central_widget)
        
        # Create information label
        self.info_label = QLabel("Click anywhere in the 3D view to test for intersection with the mesh")
        self.info_label.setAlignment(Qt.AlignCenter)
        layout.addWidget(self.info_label)
        
        # Create the 3D view
        self.view = gl.GLViewWidget()
        layout.addWidget(self.view, 1)
        
        # Configure the view
        self.view.setCameraPosition(distance=20, elevation=30, azimuth=45)
        
        # Add coordinate system
        self.view.addItem(gl.GLAxisItem())
        
        # Create and add the mesh (a plane similar to GridItem)
        self.mesh_item = self.create_plane_mesh()
        self.view.addItem(self.mesh_item)
        
        # Create and add a point item to visualize the test point
        self.point_item = gl.GLScatterPlotItem(
            pos=np.array([[0, 0, 0]]), 
            color=(1, 0, 0, 1), 
            size=0.5, 
            pxMode=False
        )
        self.view.addItem(self.point_item)
        
        # Add a grid for reference
        grid = gl.GLGridItem()
        grid.setSize(10, 10, 10)
        grid.setSpacing(1, 1, 1)
        self.view.addItem(grid)
        
        # Connect mouse click event
        self.view.mousePressEvent = self.handle_mouse_click

    def create_plane_mesh(self):
        """Create a plane mesh similar to GridItem but with solid triangles"""
        # Create vertices for a plane
        x = np.linspace(-5, 5, 20)
        y = np.linspace(-5, 5, 20)
        z = np.zeros(20)
        
        verts = []
        faces = []
        
        # Create vertices and faces
        for i in range(len(x) - 1):
            for j in range(len(y) - 1):
                # Define four vertices for a quad
                v1 = [x[i], y[j], 0]
                v2 = [x[i+1], y[j], 0]
                v3 = [x[i+1], y[j+1], 0]
                v4 = [x[i], y[j+1], 0]
                
                # Add vertices to list
                start_idx = len(verts)
                verts.extend([v1, v2, v3, v4])
                
                # Create two triangles for each quad
                faces.append([start_idx, start_idx+1, start_idx+2])
                faces.append([start_idx, start_idx+2, start_idx+3])
        
        # Convert to numpy arrays
        verts = np.array(verts, dtype=np.float32)
        faces = np.array(faces, dtype=np.uint32)
        
        # Create mesh item
        mesh = gl.GLMeshItem(
            vertexes=verts, 
            faces=faces, 
            smooth=False,
            color=(0, 0.7, 1, 0.5),  # Semi-transparent blue
            drawEdges=True,
            edgeColor=(0, 0, 0, 1)    # Black edges
        )
        return mesh

    def handle_mouse_click(self, event):
        """Handle mouse clicks to test for intersection"""
        # Only handle left clicks
        if event.button() != QtCore.Qt.LeftButton:
            return
        
        # Get mouse position in view coordinates
        pos = event.pos()
        
        # Convert to 3D world coordinates
        
        
        
        world_pos = map_2D_coords_to_3D(self.view, pos.x(), pos.y())
        if world_pos is None:
            return
        
        # Create a QVector3D point
        point = pg.Vector(world_pos.x(), world_pos.y(), world_pos.z())
        
        # Update the test point visualization
        self.point_item.setData(pos=np.array([[point.x(), point.y(), point.z()]]))
        
        # Test for intersection
        intersection = self.ray_mesh_intersection(self.mesh_item, point)
        
        # Update the info label
        if intersection:
            self.info_label.setText(
                f"Intersection found at: ({intersection.x():.2f}, {intersection.y():.2f}, {intersection.z():.2f})"
            )
            self.info_label.setStyleSheet("color: green; font-weight: bold;")
        else:
            self.info_label.setText("No intersection found")
            self.info_label.setStyleSheet("color: red; font-weight: bold;")
        
        # Update the view
        self.view.update()
        
        # Call the original mousePressEvent to handle camera rotation
        super(gl.GLViewWidget, self.view).mousePressEvent(event)

    def ray_mesh_intersection(self, mesh_item, point, direction=None):
        """
        Check if a ray from point in direction intersects with the mesh.
        Returns intersection point or None.
        """
        # Default direction is downward (negative Z)
        if direction is None:
            direction = pg.Vector(0, 0, -1)
        
        # Get mesh data
        md = mesh_item.meshData
        vertices = md.vertexes()
        faces = md.faces()
        
        closest_intersection = None
        min_distance = float('inf')
        
        # Check each triangle
        for face in faces:
            v0 = pg.Vector(*vertices[face[0]])
            v1 = pg.Vector(*vertices[face[1]])
            v2 = pg.Vector(*vertices[face[2]])
            
            # Apply mesh transformation to vertices
            transform = mesh_item.transform()
            v0 = transform.map(v0)
            v1 = transform.map(v1)
            v2 = transform.map(v2)
            
            # Ray-triangle intersection test
            intersection = self.ray_triangle_intersection(point, direction, v0, v1, v2)
            if intersection:
                # Calculate distance to the intersection point
                dist = (intersection - point).length()
                if dist < min_distance:
                    min_distance = dist
                    closest_intersection = intersection
                    
        return closest_intersection

    def ray_triangle_intersection(self, origin, direction, v0, v1, v2):
        """Möller–Trumbore ray-triangle intersection algorithm"""
        epsilon = 1e-6
        
        edge1 = v1 - v0
        edge2 = v2 - v0
        h = direction.cross(edge2)
        a = edge1.dot(h)
        
        if a > -epsilon and a < epsilon:
            return None  # Ray parallel to triangle
            
        f = 1.0 / a
        s = origin - v0
        u = f * s.dot(h)
        
        if u < 0.0 or u > 1.0:
            return None
            
        q = s.cross(edge1)
        v = f * direction.dot(q)
        
        if v < 0.0 or u + v > 1.0:
            return None
            
        t = f * edge2.dot(q)
        if t > epsilon:
            return origin + direction * t
        else:
            return None

if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = MeshPointIntersectionDemo()
    window.show()
    sys.exit(app.exec())


import numpy as np
import pyqtgraph as pg
import pyqtgraph.opengl as gl
from OpenGL import GL
from PySide6 import QtWidgets, QtGui, QtCore
import sys


def create_text_texture_atlas(texts, font_size=32):
    # Create a single QImage with all texts side by side
    font = QtGui.QFont("Arial", font_size)
    metrics = QtGui.QFontMetrics(font)

    paddings = 10
    widths = [metrics.horizontalAdvance(t) + paddings for t in texts]
    heights = [metrics.height() + paddings for _ in texts]
    total_width = sum(widths)
    max_height = max(heights)

    image = QtGui.QImage(total_width, max_height, QtGui.QImage.Format_RGBA8888)
    image.fill(QtCore.Qt.transparent)

    painter = QtGui.QPainter(image)
    painter.setFont(font)
    painter.setPen(QtGui.QColor(255, 255, 255))

    positions = []
    x_offset = 0
    for text, w in zip(texts, widths):
        painter.drawText(x_offset, metrics.ascent() + paddings // 2, text)
        positions.append((x_offset, w))
        x_offset += w
    painter.end()

    return image, positions, total_width, max_height


def build_label_mesh(text_positions, atlas_width, atlas_height, xyz_positions):
    # Create quads for each text label, mapped to texture atlas
    verts = []
    faces = []
    texcoords = []
    face_idx = 0

    for i, ((x_offset, width), pos) in enumerate(zip(text_positions, xyz_positions)):
        w_norm = width / atlas_width
        h_norm = 1.0

        u0 = x_offset / atlas_width
        u1 = (x_offset + width) / atlas_width
        v0 = 0.0
        v1 = 1.0

        size = 0.5  # physical size in 3D

        # Define quad in 3D (XY plane)
        verts.extend([
            pos + [0, 0, 0],
            pos + [size, 0, 0],
            pos + [size, size * 0.5, 0],
            pos + [0, size * 0.5, 0],
        ])
        faces.append([face_idx, face_idx+1, face_idx+2])
        faces.append([face_idx, face_idx+2, face_idx+3])
        texcoords.extend([
            [u0, v1],
            [u1, v1],
            [u1, v0],
            [u0, v0],
        ])
        face_idx += 4

    return np.array(verts), np.array(faces), np.array(texcoords)


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.widget = gl.GLViewWidget()
        self.setCentralWidget(self.widget)
        self.widget.opts['distance'] = 10

        texts = ["AAPL", "GOOG", "SPY"]
        positions = [np.array([-2, 0, 0]), np.array([0, 0, 0]), np.array([2, 0, 0])]

        image, atlas_positions, w, h = create_text_texture_atlas(texts)
        verts, faces, texcoords = build_label_mesh(atlas_positions, w, h, positions)

        mesh_item = gl.GLMeshItem(vertexes=verts, faces=faces, smooth=False, drawEdges=False)
        mesh_item.shader()['texCoords'] = texcoords

        # Convert QImage to OpenGL texture
        image = image.convertToFormat(QtGui.QImage.Format_RGBA8888)
        ptr = image.bits()
        data = np.array(ptr).reshape((image.height(), image.width(), 4))
        tex = GL.GLTexture2D(data)
        mesh_item.setGLOptions('additive')
        mesh_item.texture = tex
        mesh_item.texcoords = texcoords
        mesh_item.glOptions = 'additive'

        # Monkey patch paint method to bind texture
        def paint_with_texture(self):
            self.texture.bind()
            gl.GLMeshItem.paint(self)

        mesh_item.paint = paint_with_texture.__get__(mesh_item, gl.GLMeshItem)

        self.widget.addItem(mesh_item)


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    w = MainWindow()
    w.show()
    sys.exit(app.exec())



import sys
from PySide6.QtWidgets import QApplication, QMainWindow
from PySide6 import QtOpenGLWidgets
from PySide6.QtCore import Qt
from OpenGL import GL


class DotItem:
    def __init__(self, *coordinate):
        
        self.coordinate=list(coordinate)
    
    def paint(self):
        GL.glColor3f(1.0, 1.0, 1.0)
        
        GL.glBegin(GL.GL_POINTS)
        print(f"self.coordinate: {self.coordinate}")
        GL.glVertex2f(*(self.coordinate))  
        GL.glEnd()



class OpenGLWidget(QtOpenGLWidgets.QOpenGLWidget):
    def __init__(self, items):
        self.items=items
        super().__init__()

    def initializeGL(self):
        GL.glClearColor(0.1, 0.1, 0.1, 1.0)  # Dark gray background
        GL.glPointSize(10)  # Set point size to 10 pixels

    def paintGL(self):
        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
        
        # Set up orthographic projection (pixel coordinates)
        width, height = self.width(), self.height()
        GL.glMatrixMode(GL.GL_PROJECTION)
        GL.glLoadIdentity()
        GL.glOrtho(0, width, height, 0, -1, 1)  # (left, right, bottom, top, near, far)
        GL.glMatrixMode(GL.GL_MODELVIEW)
        GL.glLoadIdentity()
        
        for item in self.items:
            item.paint()
            
    def resizeGL(self, width, height):
        GL.glViewport(0, 0, width, height)


class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        
        self.setWindowTitle("OpenGL Point Example")
        self.setGeometry(100, 100, 800, 600)
        
        dots = [
            DotItem(100, 100),  # Top-left
            DotItem(400, 300),  # Center
            DotItem(700, 500),   # Bottom-right
        ]
           #     dots = [DotItem((0.1 * i, 0.0)) for i in range(-5, 5)]
        
        self.opengl_widget = OpenGLWidget(dots)
        self.setCentralWidget(self.opengl_widget)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec())


import numpy as np
from pyqtgraph.Qt import QtCore, QtGui, QtWidgets
from pyqtgraph import Transform3D
from pyqtgraph.opengl import GLLinePlotItem, GLViewWidget, GLTextItem
from pyqtgraph import SRTTransform3D

# Create parent lines (4 parallel lines)
# Create parent lines (4 parallel lines)
line_length = 1
y_offsets = np.linspace(-1, 1, 4)  # 4 horizontal lines

lines = []
for y in y_offsets:
    lines.append([0, y, 0])   # start point
    lines.append([1, y, 0])   # end point

parent_positions = np.array(lines)  # shape: (8, 3)


app = QtWidgets.QApplication([])
view = GLViewWidget()
view.show()

# Parent lines (white)
parent_lines = GLLinePlotItem(
    pos=parent_positions,
    color="white",
    width=2,
    mode='lines',
    antialias=True
)
view.addItem(parent_lines)

# Child lines (red, starts same as parent)

child_positions = parent_positions.copy()

#child_positions[:,0]/=5
child_lines = GLLinePlotItem(
    pos=child_positions,
    color="cyan",
    width=2,
    mode='lines',
    antialias=True
)
view.addItem(child_lines)

# Parent transform (rotate over time)
parent_transform = SRTTransform3D()
parent_lines.setTransform(parent_transform)

print(parent_lines.transform())
#parent_transform.rotate(30, (0, 0, 1))  # Initial rotation
print(child_lines.pos)
# Child transform (offset + scale)
child_transform = SRTTransform3D()
child_lines.setTransform(child_transform)
#child_lines.translate(-1.5, 0, 0, local=True) 
child_lines.translate(-1.5, 0, 0, local=False) 


child_transform.update()
#child_lines.setData(pos=child_lines.mapToParent(parent_lines.pos.T).T)

#child_transform.scale(0.5, 0.5, 0.5)  # Scale down

pos = np.array([0, 0, 0])
colors=["white", "yellow", "cyan"]
for k in range(3):
    
    pos = np.zeros(3)
    pos[k] = 1 

    
    for i in range(21):
        if i % 2 == 0:
            gl_item = GLTextItem(pos=pos * i / 10, text=f"{i}", color=colors[k])
            view.addItem(gl_item), #ignoreBounds=True)
        pos = np.zeros(3)
        pos[k] = 1 
        



def update():

    # Rotate parent
    parent_lines.transform().translate(0.5, 0, 0)
    parent_lines.setData(pos=parent_lines.transform().map(parent_positions.T).T)
    
    # Apply parent + child transforms to child
    #combined_transform = SRTTransform3D(parent_transform)
    #combined_transform.translate(*child_transform.getTranslation())
    #combined_transform.scale(*child_transform.getScale())
    
    
    #combined_transform = Transform3D(parent_lines.transform())
   # combined_transform * child_lines.transform()
    child_lines.transform().translate(0.5, 0, 0)
    transformed_child_positions = child_lines.mapToParent(child_positions.T ).T
    child_lines.setData(pos=transformed_child_positions)
# Animation timer
timer = QtCore.QTimer()
timer.timeout.connect(update)
timer.start(500)

if __name__ == '__main__':
    QtWidgets.QApplication.instance().exec()


from PySide6.QtWidgets import (QMainWindow, QWidget, QDockWidget, QSplitter, 
                              QVBoxLayout, QLabel, QApplication)
from PySide6.QtCore import Qt
import sys
import numpy as np
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from mpl_toolkits.mplot3d import Axes3D

class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        
        self.setWindowTitle("3D Graph with 2D Slices")
        self.setGeometry(100, 100, 1000, 800)
        
        # Create central widget with 3D graph
        self.central_widget = QWidget()
        self.setCentralWidget(self.central_widget)
        
        # Create 3D plot
        self.figure_3d = Figure()
        self.canvas_3d = FigureCanvas(self.figure_3d)
        self.ax_3d = self.figure_3d.add_subplot(111, projection='3d')
        
        # Generate some sample 3D data
        x = np.linspace(-5, 5, 50)
        y = np.linspace(-5, 5, 50)
        x, y = np.meshgrid(x, y)
        z = np.sin(np.sqrt(x**2 + y**2))
        
        # Plot the 3D surface
        self.ax_3d.plot_surface(x, y, z, cmap='viridis')
        self.ax_3d.set_title("3D Graph")
        
        # Create layout for central widget
        layout = QVBoxLayout(self.central_widget)
        layout.addWidget(self.canvas_3d)
        
        # Create dock widget for 2D slices
        self.create_slice_dock()
        
        # Connect 3D graph interaction to update slices
        # (You'll need to implement this based on your interaction method)
        
    def create_slice_dock(self):
        """Create the dock widget with split view of 2D slices"""
        self.dock = QDockWidget("2D Slices", self)
        self.dock.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
        
        # Create container widget for the dock
        container = QWidget()
        layout = QVBoxLayout(container)
        
        # Create splitter for the two slice views
        splitter = QSplitter(Qt.Vertical)
        
        # Create XY slice plot
        self.figure_xy = Figure()
        self.canvas_xy = FigureCanvas(self.figure_xy)
        self.ax_xy = self.figure_xy.add_subplot(111)
        self.ax_xy.set_title("XY Slice")
        
        # Create XZ slice plot
        self.figure_xz = Figure()
        self.canvas_xz = FigureCanvas(self.figure_xz)
        self.ax_xz = self.figure_xz.add_subplot(111)
        self.ax_xz.set_title("XZ Slice")
        
        # Add plots to splitter
        splitter.addWidget(self.canvas_xy)
        splitter.addWidget(self.canvas_xz)
        
        layout.addWidget(splitter)
        
#        container.setLayout(layout)
        
        self.dock.setWidget(container)
        self.addDockWidget(Qt.RightDockWidgetArea, self.dock)
        print(self.dock.widget().layout().itemAt(0).widget())
        # Enable dock features
        self.dock.setFeatures(QDockWidget.DockWidgetMovable | 
                             QDockWidget.DockWidgetFloatable |
                             QDockWidget.DockWidgetClosable)
        
    def update_slices(self, x, y, z):
        """Update the 2D slices based on a point in 3D space"""
        # This is a placeholder - implement your actual slice calculation here
        
        # Example: Update XY slice (constant z)
        self.ax_xy.clear()
        # ... plot XY slice data ...
        self.ax_xy.set_title(f"XY Slice at Z={z:.2f}")
        self.canvas_xy.draw()
        
        # Example: Update XZ slice (constant y)
        self.ax_xz.clear()
        # ... plot XZ slice data ...
        self.ax_xz.set_title(f"XZ Slice at Y={y:.2f}")
        self.canvas_xz.draw()

if __name__ == "__main__":
    app = QApplication(sys.argv)
    main_window = MainWindow()
    main_window.show()
    sys.exit(app.exec())


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from instruments.instruments import Option
    from instruments.utils import InstrumentManager
    from ...engines.interpolation_engines import Abstract3DInterpolator
    from ...data_classes.classes import VolatilityData, VolVector

from PySide6 import QtWidgets
from ...quantities import engines
from . import stylesheets as vol_table_utils
import numpy as np
from PySide6.QtCore import Qt
from . import settings, main_view 
from .. import abstract_classes
from ...data_classes import builders as builders_data_classes


class Interface(abstract_classes.ABCInterface):
    def __init__(self,
                 instrument_manager: InstrumentManager=None,
                 surface_calibration_config: Dict[str, Dict[str, Any]]={},
                 tick_engine_manager=None,
                 vol_vect_container=None,
                 **configs
                 ):
        super().__init__()
        volatility_data_container = builders_data_classes.create_volatility_data_from_vol_vect(vol_vect_container)
        
        self.splot_flag=True
        self.subplots_flag=True
        self.on_view=False
        self.plot_interaction_buffer=[]
        
        if tick_engine_manager is None:
            self.tick_engine_manager = engines.TickEngineManager("Strike",
                                                                 "Date",
                                                                 "Implied Volatility"
                                                                 )
        else:
            self.tick_engine_manager=tick_engine_manager

        self.data_view = main_view.MainView(instrument_manager,
                                            interpolation_config,
                                            self.tick_engine_manager)
                
        self.initSpotPrice(instrument_manager)
        
        self.widget_settings = settings.Settings(self,
                                                 self.tick_engine_manager,
                                                 self.data_view.metric_axis_engine,
                                                 self.spot_qlabel)
        layout_v = QtWidgets.QVBoxLayout()
        layout_v.addWidget(self.widget_settings)
        layout_v.addWidget(self.data_view)
        self.setLayout(layout_v)
                
    def initSpotPrice(self, instrument_manager):
        if len(instrument_manager.spot_instrument_container.objects) > 0:
            self.spot_name = list(instrument_manager.spot_instrument_container.objects.keys())[0]
            self.spot_object = list(instrument_manager.spot_instrument_container.objects.values())[0]
            
            self.spot_qlabel = QtWidgets.QLabel(f"{self.spot_name}: {self.spot_object.mid}")
            self.spot_qlabel.setStyleSheet(vol_table_utils.get_settings_stylesheets("SpotQLabel"))
            self.spot_qlabel.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        else:
            self.spot_qlabel=None

    def update_spot_text(self):
        if not self.spot_qlabel:
            self.spot_qlabel.setText(f"{self.spot_name}:  {f"{self.spot_object.mid:,.2f}"}")
    


import numpy as np
import pyqtgraph as pg
import pyqtgraph.opengl as gl
from PySide6.QtWidgets import QApplication
import sys

def main():
    pg.setConfigOptions(antialias=True)
    app = QApplication(sys.argv)

    view = gl.GLViewWidget()
    view.setWindowTitle('Manual 3D Graph with PySide6')
    view.setGeometry(100, 100, 800, 600)
    view.show()

    # Node positions
    num_nodes = 10
    pos = np.random.rand(num_nodes, 3) * 10

    # Adjacency list (edges)
    edges = np.array([
        [0, 1], [1, 2], [2, 3], [3, 4], [4, 0],
        [5, 6], [6, 7], [7, 8], [8, 9], [9, 5]
    ])

    # Plot nodes
    colors = np.ones((num_nodes, 4), dtype=float)
    colors[:, 0] = np.linspace(0.2, 1, num_nodes)
    sizes = np.linspace(5, 15, num_nodes)

    scatter = gl.GLScatterPlotItem(pos=pos, color=colors, size=sizes, pxMode=True)
    view.addItem(scatter)

    # Plot edges
    for edge in edges:
        p1 = pos[edge[0]]
        p2 = pos[edge[1]]
        pts = np.array([p1, p2])
        line = gl.GLLinePlotItem(pos=pts, color=(1, 1, 1, 0.6), width=1, antialias=True)
        view.addItem(line)

    # Add grid
    grid = gl.GLGridItem()
    grid.scale(2, 2, 1)
    view.addItem(grid)

    sys.exit(app.exec())

if __name__ == "__main__":
    main()



from PySide6 import QtWidgets, QtCore, QtGui
import sys

class OptionsTab(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        layout = QtWidgets.QVBoxLayout(self)

        self.tool_button = QtWidgets.QToolButton()
        self.tool_button.setText("Option Title")
        self.tool_button.setPopupMode(QtWidgets.QToolButton.InstantPopup)

        menu = QtWidgets.QMenu(self.tool_button)

        self.actions = []
        for i in range(5):
            action = QtGui.QAction(f"Option {i + 1}", self)
            action.setCheckable(True)
            menu.addAction(action)
            self.actions.append(action)

        self.action_group = QtGui.QActionGroup(self)
        self.action_group.setExclusive(False)
        for action in self.actions:
            self.action_group.addAction(action)

        # Set the menu to the tool button
        self.tool_button.setMenu(menu)

        # Add the tool button to the layout
        layout.addWidget(self.tool_button)

class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Dropdown Options with QToolButton")
        self.resize(400, 200)

        tabs = QtWidgets.QTabWidget()
        self.setCentralWidget(tabs)

        options_tab = OptionsTab()
        tabs.addTab(options_tab, "Options")

if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec())



from __future__ import annotations
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from pyvol_terminal.terminal.application import Application
    from pyvol_terminal.interfaces.abstract_classes import ABCInterface

from PySide6 import QtCore, QtWidgets
import time
from pyvol_terminal.settings import widgets as widgets_settings
import time

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, app: Application, **config):
        super().__init__()
        self.app=app
        interface_configs = config["interface_configs"]

        self.last_plot_update = time.time()
        self.interfaces={}
        
        interface_classes={}
        for idx, (name, inner_dict) in enumerate(interface_configs.items()):
            interface_classes[name] = inner_dict["interface_class"]
        
        self.interface_name_slot_map = {}
                
        self._initInterfaces(interface_configs, config)
        
        self.settings_main = widgets_settings.WindowSettings([self.app.open_main_window],
                                                             self.interface_name_slot_map)
        self._initLayout()
        
        self.settings_main.toggle_price_type(name=list(interface_classes.keys())[0])
        
    def _initInterfaces(self, interface_configs, extra_configs):
        self.interfaces={}
        self.stacked_interfaces = QtWidgets.QStackedLayout()
        for interface_name, interface_config in interface_configs.items():
            interface_class = interface_config["interface_class"]
            interface: ABCInterface = interface_class(parent=self, **interface_config, **extra_configs)
            self.app.surface_engine_worker.calibratedSignal.connect(interface.main_view.update_view)
            self.interfaces[interface_name] = interface
            self.stacked_interfaces.addWidget(interface)
            self.interface_name_slot_map[interface_name] = [lambda id, checked: self.stacked_interfaces.setCurrentIndex(id-1) if checked else None]
            
            
    def _initLayout(self):     
        self.setCentralWidget(QtWidgets.QWidget())
        self.central_layout = QtWidgets.QVBoxLayout()
        self.central_layout.addWidget(self.settings_main)
        self.central_layout.addLayout(self.stacked_interfaces)
        self.centralWidget().setLayout(self.central_layout)
    
    def closeEvent(self, event):
        #self.price_process_worker.update_interface_signal.disconnect(self.current_interface.update_interface)
        #loop = asyncio.get_event_loop()
        #loop.run_nutil_complete(loop.shutdown_asyncgens()) 
        event.accept()

        
class SubWindow(QtWidgets.QMainWindow):
    def __init__(self, widget, geometry: QtCore.Qsize, **kwargs):
        super().__init__(**kwargs)
        self.setCentralWidget(widget)
        self.setGeometry(geometry)
        self.show()
        
        
        
         